
#include	"js_simple.h"
#include	"rat_arith.h"

/*
 * Erdos conjecture is that 4/p = 1/a + 1/b + 1/c
 *	has solutions for every p.  It is enough to consider
 *	only primes p = 4k+1.
 * Allen conjecture is that there will always be a solution
 *	4/p = 1/mp + 1/q + whatever
 * where 1 <= m < ? and q is the smallest integer greater than
 *	(16*m*k*k + 8*m*k + m) / (16*m*k + 4*m - 4*k - 1)
 * or perhaps
 *	k / (1-1/(m*4))
 *	4*m*k / (4*m-1)
 *
 * Or use next larger q (see Jexc 1)
 * This solves all Tried!
 * Hard case = 4/2031121
 *
 */
#define	EXIT_GOOD	0

int show3(Ratio target, int only)
{
	double	x, x2;
	int	startj, j, j2, found = 0;
	Ratio	sub1, sub2, res1, res2;

	x = target.den / (double)target.num;
	startj = x + 1;
	for (j = startj; j <= 3*x + 0.00001; j++) {
		if (only && j >= (int)(x + 2))
			break;
		sub1 = set_rat(1, j);
		res1 = sub_rat(target, sub1);
		if (res1.num == 1) {
			found++;
			continue;
		}
		x2 = j*x / (j - x);
		for (j2 = j; j2 <= 2*x2 + 0.00001; j2++) {
			sub2 = set_rat(1, j2);
			res2 = sub_rat(res1, sub2);
			if (res2.num == 1) {
				printf("%s = %s + %s + %s\n",
					pri_rat(target),
					pri_rat(sub1),
					pri_rat(sub2),
					pri_rat(res2));
				found++;
			}
		}
	}
	return !!found;
}

int show2(Ratio target, int only)
{
	double	x;
	int	startj, j, found = 0;
	Ratio	sub1, res;

	x = target.den / (double)target.num;
	startj = x + 1;
	for (j = startj; j <= startj + startj; j++) {
		if (only && j >= startj + only)
			break;
		sub1 = set_rat(1, j);
		res = sub_rat(target, sub1);
		if (res.num == 1) {
			found++;
			printf("%s = %s + %s (Jexc %d)\n",
				pri_rat(target),
				pri_rat(sub1),
				pri_rat(res),
				j - startj);
		}
	}
	return !!found;
}

int main(int argc, char **argv)
{
	int	p, m, found;
	Ratio	target, targ2, tsub;

	setlinebuf(stdout);

#if	0
/*
No nice solution with p = 73
No nice solution with p = 193
No nice solution with p = 241
No nice solution with p = 313
No nice solution with p = 409
No nice solution with p = 433
No nice solution with p = 601
No nice solution with p = 673
No nice solution with p = 769
No nice solution with p = 1033
No nice solution with p = 1129	*
No nice solution with p = 1201	*
No nice solution with p = 1249
No nice solution with p = 1321
No nice solution with p = 1489
No nice solution with p = 1609
No nice solution with p = 1753
No nice solution with p = 1873
No nice solution with p = 1993
No nice solution with p = 2089
No nice solution with p = 2161
No nice solution with p = 2281
No nice solution with p = 2473
No nice solution with p = 2521	*
No nice solution with p = 2689	*
No nice solution with p = 2713
No nice solution with p = 2833
No nice solution with p = 2953
No nice solution with p = 3001
* - No solution for j+1 either.
*/

	// for (p = 3; p < 3557; p = nxtprime(p)) {
	for (p = 3; p < 99999; p = nxtprime(p)) {
		target = set_rat(4, p);
		found = show3(target, 1);
		if (!found)
			printf("No nice solution with p = %d\n", p);
	}
	exit(EXIT_GOOD);
#endif

	for (p = 3; p < 9999999; p = nxtprime(p))
	if (p % 4 == 1) {
		target = set_rat(4, p);
#if	1
		found = 0;
		for (m = 1; m < 100*p; m++) {
			tsub = set_rat(1, p*m);
			targ2 = sub_rat(target, tsub);
			found = show2(targ2, 3);
			if (found) {
				printf("M = %d Solved %s %s\n",
						m,
						pri_rat(target),
						pri_rat(targ2));
				break;
			}
		}
		if (!found) {
			printf("M = %d SO SAD about %s\n",
					m, pri_rat(target));
		}
#else
		found = show2(target, 0);
		found |= show3(target, 0);
		if (!found)
			printf("Erdos conjecture fails at p = %d\n", p);
#endif
	}
	exit(EXIT_GOOD);
}
