
/*
 * Add HISTO mode!!  Read input histogram Walker.
 * Add more distributions!
 *   Gamma
 *
 * To-Do:
 *   Add a more direct rand_binomial to j_random.c
 */
#include	<stdlib.h>
#include	<stdio.h>
#include	<time.h>
#include	<math.h>
#include	<string.h>
#include	"j_random.h"
/* Next is all we use from simplesubs.o, so don't bother include js_simple.h */
void *mustmalloc(size_t siz, char *msg);

#define	EXIT_GOOD	0
#define	EXIT_PINCON	25	/* arguments inconsistent */
#define	EXIT_USAGE	64	/* command line usage error */

double	Parm1 = 0, Parm2 = 0;
int	DParm1 = 0, DParm2 = 1;
int	P1_set, P2_set, Verbose;
int	Numtrials = 2000000000;
int	Method, Randseed = -666;
int	Addend = 1;
int	BMethDim;

/* Methods */
#define	M_DICE	1
#define	M_BINOM	2
#define	M_POISS	3
#define	M_GEOM	4
#define	M_HWALK	6
#define	M_GAUSS	101
#define	M_UNIFO	102
#define	M_CAUCH	103
#define	M_B3I	104
#define	M_B3S	105
#define	M_EXPON	106
#define	M_WEIBU	107
#define	M_AEPP	108

#define	MAX_HSIZ 15000
struct walkerent Histwalk[MAX_HSIZ];
char	*Strarr[MAX_HSIZ];

int main(int argc, char **argv)
{
	int	isum, j;
	int	tot, cnt;
	double	Coo[12];	/* coordinates for point in N-space */
	char	Buff[10000], walkv[1000];
	int	hwsiz, wcnt, wsss;
	double	harriv, hf;
	int	hnum;

	while (argc > 1) {
		switch (argv[1][0]) {
		case 'G': case 'g':
			if (argv[1][1] == 'E' || argv[1][1] == 'e')
				Method = M_GEOM;
			else
				Method = M_GAUSS;
			break;
		case 'E': case 'e':
			Method = M_EXPON;
			break;
		case 'A': case 'a':
			Method = M_AEPP;
			break;
		case 'W': case 'w':
			Method = M_WEIBU;
			break;
		case 'U': case 'u':
			Method = M_UNIFO;
			break;
		case 'H': case 'h':
			Method = M_HWALK;
			break;
		case 'C': case 'c':
			Method = M_CAUCH;
			break;
		case 'D': case 'd':
			Method = M_DICE;
			break;
		case 'B': case 'b':
			if ('1' < argv[1][1] && argv[1][1] <= '9') {
				if (argv[1][2] == 'S' || argv[1][2] == 's')
					Method = M_B3S;
				else
					Method = M_B3I;
				BMethDim = argv[1][1] - '0';
				break;
			}
			Method = M_BINOM;
			break;
		case 'P': case 'p':
			Method = M_POISS;
			break;
		case '-':
			if (argv[1][1] == 'n') {
				if (--argc == 1)
					goto usage;
				Numtrials = atoi(argv++[2]);
				break;
			} else if (argv[1][1] == 's') {
				if (--argc == 1)
					goto usage;
				Randseed = atoi(argv++[2]);
				break;
			} else if (argv[1][1] == 'v') {
				Verbose = 1;
				break;
			} else if (argv[1][1] == 'z') {
				Addend = 0;
				break;
			}
			/* FALL THRU */
		case '0': case '1': case '2': case '3':
		case '4': case '5': case '6': case '7':
		case '8': case '9': case '.':
			if (P2_set) {
				goto usage;
			} else if (P1_set) {
				Parm2 = atof(argv[1]);
				DParm2 = atoi(argv[1]);
				P2_set = 1;
			} else {
				Parm1 = atof(argv[1]);
				DParm1 = atoi(argv[1]);
				P1_set = 1;
			}
			break;
		default:
			goto usage;
		}
		argc--;
		argv++;
	}
	if (!Method || !P1_set)
		goto usage;

	rand_seed(Randseed == -666 ? time(0) : Randseed);

	switch (Method) {
	case M_DICE:
		while (Numtrials--) {
			if (DParm2 > 1)
				for (isum = 0, j = DParm2; j; j--)
					isum += rand_index(DParm1) + Addend;
			else
				isum = rand_index(DParm1) + Addend;
			printf("%d\n", isum);
		}
		break;
	case M_BINOM:
		while (Numtrials--) {
			if (DParm2 > 1)
				for (isum = 0, j = DParm2; j; j--)
					isum += rand_decide(Parm1);
			else
				isum = rand_decide(Parm1);
			printf("%d\n", isum);
		}
		break;
	case M_POISS:
		Parm1 = exp(Parm1);
		while (Numtrials--)
			printf("%d\n", rand_poisson(Parm1));
		break;
	case M_GEOM:
		if (Parm1 >= 1) {
		badprob:
			fprintf(stderr, "Prob [#P1] must be less than 1\n");
			exit(EXIT_PINCON);
		}
		while (Numtrials--)
			printf("%d\n", rand_geomd(Parm1) - ! Addend);
		break;
	case M_GAUSS:
		if (Parm1 <= 0) {
			fprintf(stderr, "Stddev [#P1] should be > 0.  Setting to 1\n");
			Parm1 = 1;
		}
		while (Numtrials--)
			printf("%f\n", rand_gaussian(Parm2, Parm1));
		break;
	case M_EXPON:
		if (Parm1 <= 0)
			goto lambnotpos;
		while (Numtrials--)
			printf("%f\n", rand_expon(Parm1));
		break;
	case M_WEIBU:
		if (Parm1 <= 0) {
		lambnotpos:
			fprintf(stderr, "Lambda [#P1] must be positive.\n");
			exit(EXIT_PINCON);
		}
		if (-0.0000000001 < Parm2 && Parm2 < 0.0000000001) {
			fprintf(stderr, "Setting k [#P2] to 1.\n");
			Parm2 = 1;
		}
		while (Numtrials--)
			printf("%f\n", rand_weibull(Parm1, Parm2));
		break;
	case M_CAUCH:
		if (! P2_set)
			Parm2 = 1;
		while (Numtrials--)
			printf("%f\n", Parm1 + Parm2 * rand_cauchy());
		break;
	case M_AEPP:
		if (Parm1 >= 1)
			goto badprob;
		Parm2 = exp(Parm2);
		while (Numtrials--) {
			cnt = rand_poisson(Parm2);
			for (tot = 0; cnt; cnt--)
				tot += rand_geomd(Parm1);
			printf("%d\n", tot);
		}
		break;
	case M_UNIFO:
		while (Numtrials--)
			printf("%f\n", rand_unif(Parm2, Parm1 + Parm2));
		break;
	case M_B3I:
		while (Numtrials--) {
			rand_inball(Coo, Parm1, BMethDim);
			for (j = 0; j < BMethDim; j++)
				printf("%f,%c",
					Coo[j],
					j == BMethDim - 1 ? '\n' : ' ');
		}
		break;
	case M_B3S:
		while (Numtrials--) {
			rand_surfpoint(Coo, Parm1, BMethDim);
			for (j = 0; j < BMethDim; j++)
				printf("%f,%c",
					Coo[j],
					j == BMethDim - 1 ? '\n' : ' ');
		}
		break;
	case M_HWALK:
		hwsiz = 0;
		while (gets(Buff)) {
			wsss = sscanf(Buff, "%d %s", &wcnt, walkv);
			if (wsss != 2) {
				fprintf(stderr,
		"Terminating read at mangled line *%s*\n", Buff);
				break;
			}
			Strarr[hwsiz] = mustmalloc(strlen(walkv + 1),
				"Buffer for Walker-Histogram values");
			strcpy(Strarr[hwsiz], walkv);
			Histwalk[hwsiz].x1 = hwsiz;
			Histwalk[hwsiz].prob1 = wcnt;
			if (++hwsiz >= MAX_HSIZ)
				break;
		}
		rand_setwalker(Histwalk, hwsiz);
		while (Numtrials--) {
			j = rand_walker(Histwalk, hwsiz);
			printf("%s\n", Strarr[j]);
		}
		break;
	}
	if (Verbose) {
		fprintf(stderr, "%d %f %f %d %d\n",
			Method, Parm1, Parm2, DParm1, DParm2);
	}
	exit(EXIT_GOOD);
  usage:
	fprintf(stderr, "Usage: randosim Dist [-n #N] [-s #S] [-v] [-z] #P1 [#P2]\n");
	fprintf(stderr, "  where #N is the number of trials (2 billion if omitted)\n");
	fprintf(stderr, "  and #S is a random seed (use time-of-day if omitted)\n");
	fprintf(stderr, "  and Dist #P1 specifies which distribution with which parameter.\n");
	fprintf(stderr, "  #P2, if omitted, defaults to Zero except as noted.\n");
	fprintf(stderr, "  '-v' sets a slightly verbose mode.\n");
	fprintf(stderr, "  '-z' makes minimum Dice and Geom zero instead of one.\n");
	fprintf(stderr, "Supported Continuous Distributions include:\n");
	fprintf(stderr, "  Gaus = Gaussian of std-dev P1 and mean P2\n");
	fprintf(stderr, "  Unif = Uniform with minimum P2 and maximum P1+P2\n");
	fprintf(stderr, "  Expon = Exponential with parameeter P1\n");
	fprintf(stderr, "  Weibull = Weibull with lambda=P1 k=P2 [default 1]\n");
	fprintf(stderr, "  Cauchy = Cauchy with mean P1 [P2 stretches]\n");

	fprintf(stderr, "  B3surf = Point on surface of 3-ball with radius P1\n");
	fprintf(stderr, "  B3interior = Point on interior of 3-ball with radius P1\n");
	fprintf(stderr, "  B9surf = Point on surface of 9-ball with radius P1\n");
	fprintf(stderr, "  B9interior = Point on interior of 9-ball with radius P1\n");
	fprintf(stderr, "     B2, B4, B5, B6, B7, B8 also work\n");

	fprintf(stderr, "Supported Discrete Distributions include:\n");
	fprintf(stderr, "  Geom = Geometric with prob P1\n");
	fprintf(stderr, "  Binom = Binomial with prob P1, summed over P2 [default 1] events\n");
	fprintf(stderr, "  Dice = P1-sided fair die, summed over P2 [default 1] events\n");
	fprintf(stderr, "      (for example, 'randosim Dice 6 2' is a crap roller)\n");
	fprintf(stderr, "  Poiss = Poisson with mean and variance P1\n");
	fprintf(stderr, "  Aeppli = Sum of N Geometric [P1] clusters where N is Poisson [P2]\n");
	fprintf(stderr, "     This is the Polya-Aeppli distribution, aka Geometric-Poisson\n");
	fprintf(stderr, "  Histog 1 = Output random tokens per histogram on stdin\n");
	fprintf(stderr, "(Only need the first 1 or 2 [or 3 for 'B3...'] of Method argument)\n");
	exit(EXIT_USAGE);
}
