
/*
 * acquire.c  --  Let one person play a game of 4-handed partnership Acquire.
 * Copyright (c) 2006 by James Dow Allen
 *
 *  This program can be distributed or modified subject to
 *  the terms of the GNU General Public License (version 2
 *  or, at your option, any later version) as published by
 *  the Free Software Foundation.
 *
 * After compiling, run in X Windows with following csh script
 *   #
 *   touch $1
 *   vi $1
 *   (cat $1 ; tee -a $1) | acquire -x
 * The argument to the script is an arbitrary unique name for the
 * game file.  First time, enter arbitrary random seed as first
 * line of that file.  To backtrack, kill the current job, rerun
 * the script, using the vi to change or delete undesired moves
 * in the game record.
 */

#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
typedef unsigned char u_char;

#define EXIT_GOOD       0
#define EXIT_INVOKE     26      /* incorrect method invocation */
#define EXIT_USAGE      64      /* command line usage error */

unsigned int sleep(unsigned int seconds); /* from unistd.h */

void finalscore(void);
void shotiles(void);
void shoposts(void);
void name_card(char *cp, int t);
void shuffle(int tdeck[], int size);
void awardbonus(int ch);
void tagtile(int t, int p);
void absub(int ns, int aw[], int ap[], int money);
void postm(int who, int ch, int newch);
void postb(int who, int amt);
void posts(int who, int ch, int amt);
void sho_tset(int p);
void playgame(void);
void magictile(int who);
void shoboard(int P, int hisx, int hisy);
void buysub(int who, int ch);
int getcorp(void);
int price(int cno, int siz);
int mworder(int who, int n, int flag, int nn[]);
int typeplay(int t);
int getnumber(void);
int shavail(int ch);
int getresp(void);
int gcsub(int c);
int gettile(void);

int	Replay = -1;
FILE	*ftty;

#define	BOARDWID	12
#define	BOARDHEI	9
#define	NUMTILE		(BOARDHEI*BOARDWID)
#define	NUMCHAIN	7
#define	SAFESIZE	11
#define	NUMPLAYER	4
#define	HANDSIZ		6	/* number of tiles in hand */
#define	TOTSHARE	25
#define	MAXBUY		3
#define	MAGICTILE	99
#define STARTCASH	6000
#define MAXGETS		800	/* adequate not preposterous */

#define	ttox(t)		((t) & 15)
#define	ttoy(t)		((t) >> 4)
#define	xytot(x, y)	((x) | (y) << 4)
#define	teqxy(t, x, y)	((t) == xytot(x, y))

void jleep(int k)
{
	if (Replay != 0)
		sleep(k);
}

char	Board[BOARDHEI+2][BOARDWID+2];
char	Chainsiz[NUMCHAIN+1];
int	Numchain;
int	Moveno;
int	Consecpass = 0;
char	rbuffer[100];
int	deckcut, deck[NUMTILE];

struct playinfo {
	int	cash;
	short	numshare[NUMCHAIN+1];
	short	tile[HANDSIZ];
	short	shodtile[HANDSIZ];
} Pinfo[NUMPLAYER];

	/* The two cpability lists you are about to see are
	 * almost identical.  You gotta problem with that?
	 */
char	*Standout = "\033[7m";
char	*Standok  = "\033[27m";
char	*Clearscr  = "\033[H\033[J";
char	*term_md  = "\033[1m";		/* Bold (Dim) */
char	*term_mr  = "\033[7m";		/* Dim */
char	*term_mb  = "\033[5m";		/* Blink (bold) */
char	*term_me  = "\033[m";		/* Cancel above */

char	*XStandout = "\033[7m";
char	*XStandok  = "\033[m";
char	*XClearscr  = "\033[H\033[2J";
char	*Xterm_md  = "\033[1m";
char	*Xterm_mr  = "\033[7m";
char	*Xterm_mb  = "\033[5m";		/* Not assigned per termcap */
char	*Xterm_me  = "\033[m";

char	*A_null  = "";

#define	BLOT		99
/* We rely on next two being equal to each other and 0 */
#define	EMPTY		0
#define	BORDER		0

char	*Playername[NUMPLAYER] = {
	"Bill", "Chas", "Glen", "Rick",
};

char	*corpname[NUMCHAIN + 1] = {
#if	1
	"unused",
	"Luxor", "Tower",
	"Ameri", "Festi", "World",
	"Conti", "Imper",
#else
	"unused",
	"Luxor", "Tower",
	"American", "Festival", "Worldwide",
	"Continental", "Imperial",
#endif
};

char	*tpname[] = {	/* See typeplay() */
	"no play ",
	"blot    ",
	"addgroup",
	"Founder ",
	"Merger  ",
	"Temp ill",
	"Perm ill",
};

int mwcomp(const void *p1, const void *p2)
{
	return Chainsiz[*(int *)p2] - Chainsiz[*(int *)p1];
}

/********** Routines to dialog with player **********/
/* Found a new company; return which company  */
int foundit(int who)
{
	int	ch;

	do {
		printf("%s: Select new company to found: ", Playername[who]);
		ch = getcorp();
	} while (ch == 0 || Chainsiz[ch]);
	if (Chainsiz[ch] || ch > NUMCHAIN) {
		printf("Illegal %d'th chain\n", NUMCHAIN + 1);
		exit(EXIT_INVOKE);
	}
	++Numchain;
	postm(who, ch, 0);
	return ch;
}

/* Merge companies; return survivor */
int mergewin(int nc, int nn[], int who)
{
	int	i, j;

	qsort(nn, nc, sizeof(*nn), mwcomp);
	for (i = 0; i < nc - 1; i++) {
		for (j = i + 1; j < nc; j++)
			if (Chainsiz[nn[i]] != Chainsiz[nn[j]])
				break;
		j -= i;
		if (j > 1)
			mworder(who, j, i, nn + i);
	}
	printf("Merging ");
	for (i = 1; i < nc; i++)
		printf(" %c (%d)", corpname[nn[i]][0], Chainsiz[nn[i]]);
	printf(" into %c (%d)\n", corpname[nn[0]][0], Chainsiz[nn[0]]);
	Numchain -= nc - 1;
	return nn[0];
}

int mworder(int who, int n, int flag, int nn[])
{
	int	t, ch, i;

	redo:
	printf("%s: Enter group to %s (", Playername[who],
			flag ? "merge next" : "survive");
	for (i = 0; i < n; i++)
		printf("%c", corpname[nn[i]][0]);
	printf("):");
	ch = getcorp();
	for (i = 0; i < n; i++)
		if (ch == nn[i])
			break;
	if (i == n) {
		printf("Repeat ");
		goto redo;
	}
	t = nn[0];
	nn[0] = nn[i];
	nn[i] = t;
}

/* Select a tile to play; return its index */
int seltile(int who)
{
	int	lplay, numleg, i, t, typ, legal[HANDSIZ];
	char	cardname[4];

	printf("%s's choice of tiles:\n", Playername[who]);
	for (numleg = i = 0; i < HANDSIZ; i++) {
		t = Pinfo[who].tile[i];
		typ = typeplay(t);
		name_card(cardname, t);
		printf("%d) %s%s%s -%s  ",
			i,
			Pinfo[who].shodtile[i] ? "" : Standout,
			cardname,
			Pinfo[who].shodtile[i] ? "" : Standok,
			tpname[typ]);
		Pinfo[who].shodtile[i] = 1;
		if (i == 2 || i == HANDSIZ-1)
			printf("\n");
		if (typ == 6)
			Pinfo[who].tile[i] = -1;
		legal[i] = typ > 4 || typ == 0 ? 0 : 1;
		if (legal[i])
			lplay = i, ++numleg;
	}
	if (numleg == 0) {
		printf("No legal plays\n");
		jleep(2);
		return HANDSIZ;
	}
	if (numleg == 1) {
		printf("Only one legal play\n");
		jleep(2);
		return lplay;
	}
	do {
		printf("%s: Enter tile 0-%d: ", Playername[who], HANDSIZ - 1);
		lplay = getnumber();
#ifdef	CHEAT
		if (lplay == MAGICTILE)
			break;
#endif
	} while (lplay < 0 || lplay >= HANDSIZ || !legal[lplay]);
	return lplay;
}

/* Buy stock */
void buysome(int who)
{
	int	val, i, ch, cheap;

	if (Numchain == 0)
		return;
	for (cheap = 9999900, ch = 1; ch <= NUMCHAIN; ch++) {
		val = price(ch, Chainsiz[ch]);
		if (val && val < cheap && shavail(ch))
			cheap = val;
	}
	if (cheap > Pinfo[who].cash)
		return;
	printf("%s: Enter shares to buy: ", Playername[who]);
	getresp();
	for (i = 0; i < MAXBUY; i++)
		buysub(who, rbuffer[i]);
}

void buysub(int who, int ch)
{
	int	val;

	ch = gcsub(ch);
	if (ch < 1)
		return;
	val = price(ch, Chainsiz[ch]);
	if (val && val <= Pinfo[who].cash && shavail(ch)) {
		Pinfo[who].cash -= val;
		Pinfo[who].numshare[ch] += 1;
		posts(who, ch, -val);
	}
}

/* Can we Declare the game over? */
int game_can_end(void)
{
	int	siz, ch, n41, n10;

	/* Is the game over? */
	for (n41 = n10 = 0, ch = 1; ch <= NUMCHAIN; ch++) {
		siz = Chainsiz[ch];
		if (siz > 40) ++n41;
		if (siz && siz < SAFESIZE) ++n10;
	}
	if (Consecpass < NUMPLAYER && n41 == 0 && (n10 || Numchain == 0))
		return 0;
	return 1;
}

/* Declare the game over; return 1 if it is */
int declared_over(int who)
{
	int	resp;
	/* The game can be declared over.  Ask who if he wants to */
	printf("%s: enter 'y' to end the game: ", Playername[who]);
	resp = 0x20 | getresp();
	if (resp != 'y')
		return 0;
	printf("%s declares the game over.\n", Playername[who]);
	return 1;
}

int shavail(int ch)
{
	int	p, sum = TOTSHARE;

	for (p = 0; p < NUMPLAYER; p++)
		sum -= Pinfo[p].numshare[ch];
	return sum;
}

void dssub(int oldch, int newch, int who, int val)
{
	int	tval, ns, maxt, n;

	ns = Pinfo[who].numshare[oldch];
	if (ns == 0)
		return;
	maxt = shavail(newch);
	if (maxt > ns / 2)
		maxt = ns / 2;
	shoboard(who, -1, 0);
	while (maxt) {
		printf(
	"%s: enter #(0,2,...,%d) shares of %s to trade for %s: ",
				Playername[who], maxt * 2,
				corpname[oldch],
				corpname[newch]);
		n = getnumber();
		if (n >= 0 && n <= maxt*2 && !(n & 1)) {
			maxt = 0;
			Pinfo[who].numshare[oldch] -= n;
			Pinfo[who].numshare[newch] += n / 2;
			tval = price(newch, Chainsiz[newch]) * n / 2;
			posts(who, oldch, tval);
			posts(who, newch, -tval);
			ns -= n;
		}
	}
	if (ns) while (1) {
		printf("%s: enter #(0,1,...,%d) shares of %s to sell: ",
				Playername[who], ns,
				corpname[oldch]);
		n = getnumber();
		if (n >= 0 && n <= ns) {
			Pinfo[who].cash += n * val;
			Pinfo[who].numshare[oldch] -= n;
			posts(who, oldch, n * val);
			break;
		}
	}
}

int getresp(void)
{
	char	dbuffer[100];

	do {
		gets(rbuffer);
		if (Replay >= 0) {
			printf("%s\n", rbuffer);
			if (Replay == 2)
				fgets(dbuffer, 100, ftty);
			else if (Replay == 1)
				jleep(1);
		}
	} while (rbuffer[0] == '#');
	strcat(rbuffer, "     ");
	return rbuffer[0];
}

int gcsub(int c)
{
	switch (c | 0x20) {
	case 'n': case ' ': return 0;
	case 'l': return 1;
	case 't': return 2;
	case 'a': return 3;
	case 'f': return 4;
	case 'w': return 5;
	case 'c': return 6;
	case 'i': return 7;
	default: return -1;
	}
}

int getcorp(void)
{
	int	ch;

	do {
		ch = gcsub(getresp());
		if (ch < 0)
			printf("No, choose from T,L,A,F,W,C,I (or N) : ");
	} while (ch < 0);
	return ch;
}

int getnumber(void)
{
	getresp();
	return atoi(rbuffer);
}

void disposeshare(int oldch, int newch, int who, int siz)
{
	int	p, val;

	val = price(oldch, siz);
	p = who;
	do {
		dssub(oldch, newch, p, val);
		if (++p >= NUMPLAYER)
			p = 0;
	} while (p != who);
}

int btoc(int x)
{
	switch (x) {
	case BLOT: return 'X';
	case EMPTY: return '.';
	}
	return corpname[x][0];
}

void renumchain(int old, int new)
{
	int	x, y;

	for (y = 1; y <= BOARDHEI; y++)
	for (x = 1; x <= BOARDWID; x++)
		if (Board[y][x] == old)
			Board[y][x] = new;
}

void play(int t, int who)
{
	int	numb, ch, numc, i, x, y, cnei[5], nei[4];

	if (t == 0)
		return;
	x = ttox(t);
	y = ttoy(t);
	nei[0] = Board[y+1][x];
	nei[1] = Board[y-1][x];
	nei[2] = Board[y][x+1];
	nei[3] = Board[y][x-1];
	for (numb = numc = i = 0; i < 4; i++)
		if (nei[i] == BLOT)
			++numb;
		else if (nei[i] && (numc == 0 || nei[i] != cnei[0] &&
				   (numc == 1 || nei[i] != cnei[1] &&
				   (numc == 2 || nei[i] != cnei[2] &&
				   (numc == 3 || nei[i] != cnei[3])))))
			cnei[numc++] = nei[i];

	if (numc == 1) {
		/* Grow a chain */
		ch = cnei[0];
		Chainsiz[ch] += 1 + numb;
		setblots:
		Board[y][x] = ch;
		if (Board[y+1][x]) Board[y+1][x] = ch;
		if (Board[y-1][x]) Board[y-1][x] = ch;
		if (Board[y][x+1]) Board[y][x+1] = ch;
		if (Board[y][x-1]) Board[y][x-1] = ch;
	} else if (numc > 1) {
		/* Merge two (or more) chains */
		cnei[numc] = 0;	/* Special kluge */
		ch = mergewin(numc, cnei, who);
		/*
		 * We'd like to do all the awardbonus(); pause
		 * with a fresh screen showing the bonuses; and then
		 * do the disposeshare()s.
		 * However the postm()/absub() for the game summary
		 * requires that the companies be handled one-by-one.
		 * Tangled web.
		 */
		jleep(3);
		for (i = 1; i < numc; i++) {
			postm(who, cnei[i], ch);
			awardbonus(cnei[i]);
			disposeshare(cnei[i], ch, who, Chainsiz[cnei[i]]);
			Chainsiz[ch] += Chainsiz[cnei[i]];
			Chainsiz[cnei[i]] = 0;
			renumchain(cnei[i], ch);
		}
		Chainsiz[ch] += 1 + numb;
		goto setblots;
	} else if (numb > 0) {
		/* Make a chain */
		ch = foundit(who);
		Chainsiz[ch] = 1 + numb;
		if (shavail(ch)) {
			/* Founder's bonus */
			Pinfo[who].numshare[ch] += 1;
		}
		goto setblots;
	} else {
		/* Lay a blot */
		Board[y][x] = BLOT;
	}
}

/*
 * Types of tile play:
 *	0 - no play
 *	1 - lay a blot
 *	2 - add to a group
 *	3 - found a group
 *	4 - merge two groups
 *	5 - found the 8'th group
 *	6 - merge off a safe group
 */
int typeplay(int t)
{
	int	num11, numb, ch, numc, i, x, y, cnei[4], nei[4];

	if (t == 0)
		return 0;
	x = ttox(t);
	y = ttoy(t);
	nei[0] = Board[y+1][x];
	nei[1] = Board[y-1][x];
	nei[2] = Board[y][x+1];
	nei[3] = Board[y][x-1];
	for (numb = numc = i = 0; i < 4; i++)
		if (nei[i] == BLOT)
			++numb;
		else if (nei[i] && (numc == 0 || nei[i] != cnei[0] &&
				   (numc == 1 || nei[i] != cnei[1] &&
				   (numc == 2 || nei[i] != cnei[2] &&
				   (numc == 3 || nei[i] != cnei[3])))))
			cnei[numc++] = nei[i];

	if (numc == 1) {
		/* Grow a chain */
		return 2;
	} else if (numc > 1) {
		/* Merge two (or more) chains */
		for (num11 = i = 0; i < numc; i++)
			if (Chainsiz[cnei[i]] >= SAFESIZE)
				++num11;
		return num11 > 1 ? 6 : 4;
	} else if (numb > 0) {
		/* Make a chain */
		return Numchain == NUMCHAIN ? 5 : 3;
	} else {
		/* Lay a blot */
		return 1;
	}
}

int	shkeepx = -1, shkeepy = -1;

int shocell(int x, int y)
{
	char	c;

	c = btoc(Board[y][x]);
	if (c == '.' && x == shkeepx && y == shkeepy)
		c = '*';
	return c;
}

void shoboard(int P, int hisx, int hisy)
{
	int	ist, j, p, i, sc, x, y;
	short	*tilep;
	static int keepP = -17;

	if (hisx) {
		shkeepx = hisx;
		shkeepy = hisy;
	}
	printf("%s", Clearscr);
	printf("\033[H\n    ");
	for (x = 1; x <= BOARDWID; x++)
		printf("%2d", x);
	printf("       (Move #%3d)       ", Moveno);
	if (P >= 0 && hisx >= 0)
		keepP = P;
	for (p = 0; p < NUMPLAYER; p++)
		printf("%s%s%s  ",
			keepP != p ? "" : Standout,
			Playername[p],
			keepP != p ? "" : Standok);
	printf("\n");
    for (y = 1; y <= BOARDHEI; y++) {
		printf(" %c  ", y + 'A' - 1);
	if (P >= 0) {
		tilep = Pinfo[P].tile;
		for (x = 1; x <= BOARDWID; x++) {
			ist = 0;
			if (shkeepx != x || shkeepy != y)
				for (j = 0; j < HANDSIZ; j++)
					if (teqxy(tilep[j], x, y)) {
						++ist;
						break;
					}
			if (ist)
				printf(" %s%d%s", Standout,
				j, Standok);
			else
				printf(" %c", shocell(x, y));
		}
	} else {
		if (y == hisy) {
			for (x = 1; x < hisx; x++)
				printf(" %c", shocell(x, y));
			printf(" %s%c%s", Standout,
				shocell(x, y), Standok);
			for (x = hisx + 1; x <= BOARDWID; x++)
				printf(" %c", shocell(x, y));
		} else for (x = 1; x <= BOARDWID; x++)
			printf(" %c", shocell(x, y));
	}
	printf("      ");
	if (y == 1) {
		printf("   Cash:         ");
		for (p = 0; p < NUMPLAYER; p++)
			printf(" %5d", Pinfo[p].cash);
	} else if (y < 9) {
		i = y - 1;
		sc = Chainsiz[i];
		if (sc) {
#if	1
			printf("%s", corpname[i]);
#else
			printf("%s", term_md);	/* or mb for blink */
			printf("%s", corpname[i]);
			printf("%s", term_me);
#endif
			printf(" %2d %4d", sc, price(i, sc));
		} else {
			printf("%s", corpname[i]);
			printf(" %2d %4d", sc, price(i, sc));
		}
		if (price(i, sc) == 0 || shavail(i) == 0)
			printf("    ", shavail(i));
		else
#if	1
			printf("%s(%2d)%s", A_null, shavail(i), A_null);
#else
			printf("%s(%2d)%s", term_md, shavail(i), term_me);
			printf("%s%2d%s", Standout, shavail(i), Standok);
#endif
		for (p = 0; p < NUMPLAYER; p++)
			if (Pinfo[p].numshare[i]) {
				printf(" %5d", Pinfo[p].numshare[i]);
			} else
				printf("     .");
	}
	printf("\n");
    }
	for (i = 0; i < 13; i++)
		printf(\
"                                                                    \n");
	printf("\033[H\n\n\n\n\n\n\n\n\n\n\n\n");
}

void initboard(void)
{
	int	x, y, p, c, t;

	shuffle(deck, NUMTILE);
	Numchain = 0;
	Moveno = 0;
	for (x = 0; x < BOARDWID + 2; x++)
	for (y = 0; y < BOARDHEI + 2; y++)
		Board[y][x] = x == 0 || y == 0
			|| x == BOARDWID + 1 || y == BOARDHEI + 1
			? BORDER : EMPTY;
	for (c = 0; c <= NUMCHAIN; c++) {
		Chainsiz[c] = 0;
	}
	for (p = 0; p < NUMPLAYER; p++) {
		for (c = 0; c <= NUMCHAIN; c++)
			Pinfo[p].numshare[c] = 0;
		Pinfo[p].cash = STARTCASH;
		for (t = 0; t < HANDSIZ; t++) {
			x = Pinfo[p].tile[t] = gettile();
			Pinfo[p].shodtile[t] = 1;
			tagtile(x, p);
		}
	}
}

int price(int cno, int siz)
{
	int	pri = (cno / 3) * 100;

	if (siz < 2)
		return 0;
	if (siz > 40)
		return pri + 1000;
	if (siz > 30)
		return pri + 900;
	if (siz > 20)
		return pri + 800;
	if (siz > 10)
		return pri + 700;
	if (siz > 5)
		return pri + 600;
	else
		return pri + siz * 100;
}

int main(int argc, char **argv)
{
	int	rseed;

	if (argv[1] && !strcmp(argv[1], "-x")) {
		Standout = XStandout;
		Standok = XStandok;
		Clearscr = XClearscr;
		term_md = Xterm_md;
		term_mr = Xterm_mr;
		term_mb = Xterm_mb;
		term_me = Xterm_me;
		argc -= 1;
		argv += 1;
	}
	if (argv[1] && argv[2] && !strcmp(argv[1], "-R")) {
		Replay = atoi(argv[2]);
		argc -= 2;
		argv += 2;
	}
	if (argv[1]) {
		printf("Usage: play_acq [-x] [-R #]\n");
		printf(" -x for xterm\n");
		printf(" -R 0 for fast playback\n");
		printf(" -R 1 for slow playback\n");
		printf(" -R 2 for interactive playback\n");
	}
	if (Replay == 2)
		ftty = fopen("/dev/tty", "r");
	setbuf(stdout, 0);
	printf("Enter Rseed: ");
	rseed = getnumber();
	srandom(rseed);
	playgame();
	exit(EXIT_GOOD);
}

void playgame(void)
{
	int	x, p, t, i, done;

	initboard();
	shoboard(-1, 0, 0);
	for (t = done = 0; ! done; ) {
		for (p = 0; p < NUMPLAYER && !done; p++) {
			++Moveno;
			redotile:
			shoboard(p, ttox(t), ttoy(t));
			i = seltile(p);
			if (i == MAGICTILE) {
				magictile(p);
				goto redotile;
			}
			t = i < HANDSIZ ? Pinfo[p].tile[i] : 0;
			play(t, p);
			shoboard(-1, ttox(t), ttoy(t));
			if (i < HANDSIZ) {
				Pinfo[p].tile[i] = -1;
				Consecpass = 0;
			} else
				++Consecpass;
			buysome(p);
			if (game_can_end()) {
				shoboard(-1, ttox(t), ttoy(t));
				done = declared_over(p);
			}
			for (i = 0; i < HANDSIZ; i++)
				if (Pinfo[p].tile[i] == -1) {
					x = Pinfo[p].tile[i] = gettile();
					Pinfo[p].shodtile[i] = 0;
					tagtile(x, p);
				}
		}
	}
	shoboard(-1, 0, 0);
	finalscore();
	jleep(4);
	shoboard(-1, 0, 0);
	printf("Game over.\n");
	jleep(4);
	shoposts();
	jleep(4);
	shotiles();
	for (p = 0; p < NUMPLAYER; p++) {
		jleep(7);
		sho_tset(p);
	}
}

void finalscore(void)
{
	int	ch, val, siz, p;

	for (ch = 1; ch <= NUMCHAIN; ch++)
	if (siz = Chainsiz[ch]) {
		val = price(ch, siz);
		postm(-1, ch, -1);
		awardbonus(ch);
		for (p = 0; p < NUMPLAYER; p++) {
			Pinfo[p].cash += val * Pinfo[p].numshare[ch];
			posts(p, ch, val * Pinfo[p].numshare[ch]);
			Pinfo[p].numshare[ch] = 0;
		}
	}
}

/* Award majority/minority bonusses */
void awardbonus(int ch)
{
	int	i, win, pla, val;
	int	ap[NUMPLAYER];
	int	aw[NUMPLAYER];

	val = price(ch, Chainsiz[ch]);
	for (i = 0; i < NUMPLAYER; i++) {
		ap[i] = Pinfo[i].numshare[ch];
		aw[i] = 0;
	}
	win = ap[0];
	pla = 0;
	for (i = 1; i < NUMPLAYER; i++)
		if (ap[i] > pla)
			if (ap[i] > win)
				pla = win, win = ap[i];
			else
				pla = ap[i];
	if (pla == 0 || pla == win)
		absub(win, aw, ap, val * 15);
	else {
		absub(win, aw, ap, val * 10);
		absub(pla, aw, ap, val * 5);
	}
	printf("Bonus for %s --", corpname[ch]);
	for (i = 0; i < NUMPLAYER; i++) {
		if (aw[i])
			printf(" %s $%d", Playername[i], aw[i]);
		Pinfo[i].cash += aw[i];
		postb(i, aw[i]);
	}
	printf("\n");
}

void absub(int ns, int aw[], int ap[], int money)
{
	int	i, nt;

	for (nt = i = 0; i < NUMPLAYER; i++)
		nt += ap[i] == ns;
	money /= 100;
	money += nt - 1;
	money /= nt;
	money *= 100;
	for (nt = i = 0; i < NUMPLAYER; i++)
		if (ap[i] == ns)
			aw[i] += money;
}

void magictile(int who)
{
	int	i, j, t;

	shuffle(deck + deckcut, NUMTILE - deckcut);
	for (j = i = 0; i < HANDSIZ; i++)
		if (Pinfo[who].tile[i] > 0) {
			t = deck[NUMTILE - 1 - j];
			deck[NUMTILE - 1 - j] = Pinfo[who].tile[i];
			Pinfo[who].tile[i] = t;
			j++;
		}
}

char	Tiletab[BOARDHEI+2][BOARDWID+2];
#define	THSIZE	((3 * NUMPLAYER + BOARDHEI*BOARDWID) / NUMPLAYER)
short	Tilehist[NUMPLAYER][THSIZE];
short	Thcnt[NUMPLAYER];

void tagtile(int t, int p)
{
	Tiletab[ttoy(t)][ttox(t)] = Playername[p][0];
	if (Thcnt[p] < THSIZE)
		Tilehist[p][Thcnt[p]++] = t;
}

int locti(int t, int p)
{
	int	i;

	for (i = 0; i < Thcnt[p]; i++)
		if (Tilehist[p][i] == t)
			return i;
	return -1;
}

void sho_tset(int p)
{
	int	x, y;
	char	c;

	printf("%s", Clearscr);
	printf("\033[H\n    ");
	printf("%s's Tiles:\n\n   ", Playername[p]);
	for (x = 1; x <= BOARDWID; x++)
		printf("%2d", x);
	printf("\n");
	for (y = 1; y <= BOARDHEI; y++) {
		printf(" %c  ", y + 'A' - 1);
		for (x = 1; x <= BOARDWID; x++)
			if ((c = Tiletab[y][x]) == Playername[p][0])
				printf("%s%2d%s", Standout,
					locti(xytot(x, y), p), Standok);
			else if (c >= 'A' && c <= 'Z')
				printf(" %c", c + ' ');
			else
				printf(" %c", c);
		printf("\n");
	}
	printf("\n");
}

int gettile(void)
{
	if (deckcut < NUMTILE)
		return deck[deckcut++];
	else
		return 0;
}

void shuffle(int tdeck[], int size)
{
	int	i, j, t;
	static int new_pack = 1;

	if (new_pack) {
		for (i = 0; i < BOARDHEI+2; i++)
		for (j = 0; j < BOARDWID+2; j++)
			Tiletab[i][j] = '.';
		for (i = 0 ; i < BOARDHEI; i++)
		for (j = 0; j < BOARDWID; j++)
			deck[i + j * BOARDHEI] = xytot(j + 1, i + 1);
		new_pack = deckcut = 0;
	}
	for (i = 0; i < size; i++) {
		j = random() % (size - i) + i;
		t = tdeck[j];
		tdeck[j] = tdeck[i];
		tdeck[i] = t;
	}
}

void name_card(char *cp, int t)
{
	if (t == 0)
		strcpy(cp, "---");
	else
		sprintf(cp, "%c%02d", 'A' - 1 + ttoy(t), ttox(t));
}

/*
 * From here to EOF is just for after-game displays.
 */

int	shareprof[NUMPLAYER][NUMCHAIN+1];
int	totprof[NUMPLAYER];
int	numevent;

struct mergeevent {
	int	chain, into, playr;
	int	gain[NUMPLAYER];
} Mev[100];

void postm(int who, int ch, int newch)
{
	Mev[numevent].chain = ch;
	Mev[numevent].playr = who;
	Mev[numevent].into = newch;
	++numevent;
}

void postb(int who, int amt)
{
	Mev[numevent - 1].gain[who] = amt;
}

void posts(int who, int ch, int amt)
{
	shareprof[who][ch] += amt;
}

void shoposts(void)
{
	struct mergeevent *mp;
	int	m, p, ch;

	printf("         ");
	for (p = 0; p < NUMPLAYER; p++) {
		printf(" %s  ", Playername[p]);
	}
	printf("\n");
	for (mp = Mev, m = 0; m < numevent; m++, mp++) {
		if (mp->into == 0)
			printf("%s founded %s\n",
				Playername[mp->playr],
				corpname[mp->chain]);
		else {
			if (mp->into == -1)
				printf("%s survived till EOG\n",
					corpname[mp->chain]);
			else
				printf("%s merged %s into %s\n",
					Playername[mp->playr],
					corpname[mp->chain],
					corpname[mp->into]);
			printf("M %s: ", corpname[mp->chain]);
			for (p = 0; p < NUMPLAYER; p++) {
				printf("%6d ", mp->gain[p]);
				totprof[p] += mp->gain[p];
			}
			printf("\n");
		}
	}
	for (ch = 1; ch <= NUMCHAIN; ch++) {
			printf("S %s: ", corpname[ch]);
			for (p = 0; p < NUMPLAYER; p++) {
				printf("%6d ", shareprof[p][ch]);
				totprof[p] += shareprof[p][ch];
			}
			printf("\n");
	}
	printf("G %s: ", "Total");
	for (p = 0; p < NUMPLAYER; p++) {
		printf("%6d ", STARTCASH + totprof[p]);
	}
	printf("\n");
	printf("         %6d               %6d\n",
			STARTCASH + totprof[0] + STARTCASH + totprof[2],
			STARTCASH + totprof[1] + STARTCASH + totprof[3]);
}

void shotiles(void)
{
	int	ist, j, p, i, sc, x, y;
	short	*tilep;

	printf("%s", Clearscr);
	printf("\033[H\n    ");
	for (x = 1; x <= BOARDWID; x++)
		printf("%2d", x);
	printf("\n");
	for (y = 1; y <= BOARDHEI; y++) {
		printf(" %c  ", y + 'A' - 1);
		for (x = 1; x <= BOARDWID; x++)
			printf(" %c", Tiletab[y][x]);
		printf("\n");
	}
	printf("\n");
}
