
#include	"poker.h"
#include	<string.h>

/*
 * holdodds.c  --  Show hold'em odds.
 * Copyright (c) 2006 by James Dow Allen
 *
 *  This program can be distributed or modified subject to
 *  the terms of the GNU General Public License (version 2
 *  or, at your option, any later version) as published by
 *  the Free Software Foundation.
 *
 * Joker is "jo" (code 0xff)
 * Code assumes 32-bit int's -- sorry.
 */

#define EXIT_GOOD       0
#define EXIT_USAGE      64      /* command line usage error */

/* Following can be 1, higher values not well supported */
/* (Frankly, don't recall testing single joker thoroughly) */
#define	NUMJOKER	0
#define	DECKSIZE	(NUMSUIT*NUMRANK + NUMJOKER)

#define	DEUCE		2
#define	JOKER		15

Card	Deck[DECKSIZE];

void initdeck()
{
	int	i, j;
	Card	*dp = Deck;

	for (i = 0; i < NUMRANK; i++)
	for (j = 0; j < NUMSUIT; j++)
		*dp++ = i + DEUCE + (1 << j + 4);
	for (i = 0; i < NUMJOKER; i++)
		*dp++ = 0xf0 | JOKER;
}

/* Find best 5-hand from set of 7.
 * Do it the brute-force obvious way.
 */
int eval7(Card c0, Card c1, Card *rest)
{
	int	best, i;

	best = pokval(rest[0], rest[1], rest[2], rest[3], rest[4]);
	if ((i = pokval(c0, c1, rest[0], rest[1], rest[2])) > best) best = i;
	if ((i = pokval(c0, c1, rest[0], rest[1], rest[3])) > best) best = i;
	if ((i = pokval(c0, c1, rest[0], rest[1], rest[4])) > best) best = i;
	if ((i = pokval(c0, c1, rest[0], rest[2], rest[3])) > best) best = i;
	if ((i = pokval(c0, c1, rest[0], rest[2], rest[4])) > best) best = i;
	if ((i = pokval(c0, c1, rest[0], rest[3], rest[4])) > best) best = i;
	if ((i = pokval(c0, c1, rest[1], rest[2], rest[3])) > best) best = i;
	if ((i = pokval(c0, c1, rest[1], rest[2], rest[4])) > best) best = i;
	if ((i = pokval(c0, c1, rest[1], rest[3], rest[4])) > best) best = i;
	if ((i = pokval(c0, c1, rest[2], rest[3], rest[4])) > best) best = i;
	if ((i = pokval(c0, rest[0], rest[1], rest[2], rest[3])) > best)best=i;
	if ((i = pokval(c0, rest[0], rest[1], rest[2], rest[4])) > best)best=i;
	if ((i = pokval(c0, rest[0], rest[1], rest[3], rest[4])) > best)best=i;
	if ((i = pokval(c0, rest[0], rest[2], rest[3], rest[4])) > best)best=i;
	if ((i = pokval(c0, rest[1], rest[2], rest[3], rest[4])) > best)best=i;
	if ((i = pokval(c1, rest[0], rest[1], rest[2], rest[3])) > best)best=i;
	if ((i = pokval(c1, rest[0], rest[1], rest[2], rest[4])) > best)best=i;
	if ((i = pokval(c1, rest[0], rest[1], rest[3], rest[4])) > best)best=i;
	if ((i = pokval(c1, rest[0], rest[2], rest[3], rest[4])) > best)best=i;
	if ((i = pokval(c1, rest[1], rest[2], rest[3], rest[4])) > best)best=i;
	return best;
}

int excluded(unsigned int i, Card *a, Card *b)
{
	i = Deck[i];
	if (i != a[0] && i != a[1] && i != b[0] &&i != b[1])
		return i;
	else
		return 0;
}

long thisto[2][10];
int Verbose = 0;

int main(int argc, char **argv)
{
	Card	handa[2], handb[2], rest[5];
	int	i1, i2, i3, i4, i5;
	long	cump, awin, bwin, draw;
	unsigned int	aval, bval;
	int	Singhand = 0;

	setlinebuf(stdout);
	initdeck();
	if (argv[1] && !strcmp(argv[1], "-verbose")) {
		argc--;
		argv++;
		Verbose = 1;
	}
	if (argc == 3) {
		Singhand = 1;
	} else if (argc != 5) {
		printf("Usage: holdodds [-verbose] #A1 #A2 #B1 #B2 -- compare two hold'em hands\n");
		printf("   Or: holdodds #A1 #A2 -- consider one hold'em hand\n");
		printf("\tCards are '9S', 'KD', etc.\n");
		exit(EXIT_USAGE);
	} else {
		handb[0] = s2i(argv[3]);
		handb[1] = s2i(argv[4]);
	}
	awin = bwin = draw = 0;
	handa[0] = s2i(argv[1]);
	handa[1] = s2i(argv[2]);
	if (Singhand) {
		handb[0] = handa[0];
		handb[1] = handa[1];
	}
	for (i1 = 0; i1 < DECKSIZE; i1++)
	if (rest[0] = excluded(i1, handa, handb))
	for (i2 = i1+1; i2 < DECKSIZE; i2++)
	if (rest[1] = excluded(i2, handa, handb))
	for (i3 = i2+1; i3 < DECKSIZE; i3++)
	if (rest[2] = excluded(i3, handa, handb))
	for (i4 = i3+1; i4 < DECKSIZE; i4++)
	if (rest[3] = excluded(i4, handa, handb))
	for (i5 = i4+1; i5 < DECKSIZE; i5++)
	if (rest[4] = excluded(i5, handa, handb)) {
		aval = eval7(handa[0], handa[1], rest);
		++thisto[0][aval >> 24];
		if (Singhand == 0) {
			bval = eval7(handb[0], handb[1], rest);
			++thisto[1][bval >> 24];
			if (aval > bval) {
				++awin;
			} else if (aval < bval) {
				++bwin;
			} else {
				++draw;
			}
		} else
			++awin;
	}
	if (Singhand == 0) {
		printf("A [%s-%s] wins (%d) = %f%%\n",
				argv[1], argv[2],
				awin, awin * 100.0 / (awin + bwin + draw));
		printf("B [%s-%s] wins (%d) = %f%%\n",
				argv[3], argv[4],
				bwin, bwin * 100.0 / (awin + bwin + draw));
		if (Verbose)
		    printf("Drawn (%d) = %f%%\n",
				draw, draw * 100.0 / (awin + bwin + draw));
	}
	if (Verbose || Singhand)
	   for (i1 = 0; i1 < 2 - Singhand; i1++, printf("\n"))
	      for (cump = i2 = 0; i2 < 10; i2++)
	         if (i3 = thisto[i1][i2]) {
			printf("Player %c gets %s   \twith prob %6.3f%%   (Or better: %6.3f%%)\n",
				"AB"[i1], valnames[i2],
				i3 * 100.0 / (awin + bwin + draw),
				 100 - cump * 100.0 / (awin + bwin + draw));
			printf("Player %c gets %s   \t%d times out of %d\n",
				"AB"[i1], valnames[i2],
				i3, awin + bwin + draw);
			cump += i3;
	         }
	exit(EXIT_GOOD);
}
