
#include	"poker.h"
#include	<time.h>

/*
 * High-performance hash table managers are found in ../jdas_htab.
 * But for simple uses, a quick-and-dirty hash table is just as good.
 *
 * Define the table size and format; then include the actual function code.
 * The following five lines (after this comment),
 *      should be considered the example invocation of ezhash.c
 * struct hashe was defined in poker.h 
 */ 
#define	NUMEL	99013
#define	NUMHCAB	5	/* 1 works fine but let's test */
struct hashe Hashe[NUMHCAB][NUMEL];
#define	HT_EMPTY(np)		((np)->valu == 0)
#define	HT_MATCH(np, kkk)	((np)->keyq == (kkk))
#define	HT_SETIT(np, key)	((np)->keyq = 1 + (key) / NUMHCAB)
#include	"ezhash.c"

#define EXIT_GOOD       0
#define EXIT_USAGE      64      /* command line usage error */
#define EXIT_SOFTWARE   70      /* internal software error */

#if	1
#define	DSIZE	52
#else
#define	COUNTWAYS
#define	DSIZE	53	/* allow Joker */
#endif

int	Forbid_quads = 0;	/* Set to 7 to forbid */

#define	MAXCOUNT	2

#define	ONE	((long long)1)

/* Set all pat hands into hash table */
void set_table(void)
{
	int	c1, c2, c3, c4, c5;
	int	cards[5];
	int	val;
	unsigned long long hmask;
	struct	hashe *hp;

	for (c1 =    0; c1 < DSIZE; c1++)
	for (c2 = c1+1; c2 < DSIZE; c2++)
	for (c3 = c2+1; c3 < DSIZE; c3++)
	for (c4 = c3+1; c4 < DSIZE; c4++)
	for (c5 = c4+1; c5 < DSIZE; c5++) {
		cards[0] = CX2I(c1);
		cards[1] = CX2I(c2);
		cards[2] = CX2I(c3);
		cards[3] = CX2I(c4);
		cards[4] = CX2I(c5);
		val = pokval(cards[0], cards[1], cards[2], cards[3], cards[4]);
		val >>= 24;
		if (val != Forbid_quads) /* Do not count Quadruplets as Pat */
		if (val >= 4) {
			/* It is straight or better */
			hmask = ONE << c1 | ONE << c2 | ONE << c3 | ONE << c4 | ONE << c5;
			hp = findnode(hmask);
			if (! HT_EMPTY(hp)) {
				printf("TWICE\n");
				exit(EXIT_SOFTWARE);
			}
			HT_SETIT(hp, hmask);
			hp->valu = val;
			/* Turn this on to list all 19716 pat hands */
			if (0)
				phand("Is a pathand", cards);
		}
	}
	return;
}

/* Do crd[0...24] with usemask excluded contain count pat hands? */
int succ_mav(int *crd, int usemask, int count)
{
	int	x1, x2, x3, x4, x5, xmsk;
	unsigned long long allc1, allc2, allc3, allc4, allc5;
	int	num, numways = 0;

	for (x1 = 0; x1 < 21; x1++)
		if (!(usemask & 1 << x1))
			break;
	if (x1 >= 21) {
		printf("Impossible.\n");
		exit(EXIT_SOFTWARE);
	}
	allc1 = ONE << crd[x1];
	for (x2 = x1+1; x2 < 22; x2++)
	if (allc2 = allc1 | ONE << crd[x2], !(usemask & 1 << x2))
	for (x3 = x2+1; x3 < 23; x3++)
	if (allc3 = allc2 | ONE << crd[x3], !(usemask & 1 << x3))
	for (x4 = x3+1; x4 < 24; x4++)
	if (allc4 = allc3 | ONE << crd[x4], !(usemask & 1 << x4))
	for (x5 = x4+1; x5 < 25; x5++)
	if (allc5 = allc4 | ONE << crd[x5], !(usemask & 1 << x5)) {
		if (findnode(allc5)->valu) {
			if (count == 1) {
#ifdef	COUNTWAYS
				numways += 1;
				continue;
#else
				return 1;
#endif
			}
			xmsk = 1 << x1 | 1 << x2 | 1 << x3 | 1 << x4 | 1 << x5;
			if (num = succ_mav(crd, usemask | xmsk, count - 1)) {
#ifdef	COUNTWAYS
				numways += num;
				if (numways >= MAXCOUNT)
					return numways;
				continue;
#else
				return 1;
#endif
			}
		}
	}
	return numways;
}

int	Numsucc[50];

int	Deck[DSIZE];

int main(int argc, char **argv)
{
	int	j, ix, iter, ns, t;
	int	nsuc;

	if (argv[1] && !strcmp(argv[1], "-Noquad"))
		Forbid_quads = 7;
	srandom(time(0));
	setlinebuf(stdout);
	set_table();
	for (j = 0; j < DSIZE; j++)
		Deck[j] = j;
	printf("Setup complete\n");
	/* The game is 25-card Maverick solitaire,
	 * But we start with 24 cards and add each of the remaining 28
	 *  as the 25th, one by one.  (This is how I play Maverick solitaire:
	 *  See if I can get the five pats all 28 ways.
	 */
	iter = 0;
	for (; iter < 200; iter++) {
		shuffle(Deck, DSIZE);
		ns = 0;
		for (j = 24; j < 52; j++) {
			t = Deck[24], Deck[24] = Deck[j], Deck[j] = t;
			ns += !! succ_mav(Deck +  0, 0, 5);
		}
		printf("%d\n", ns);
		if (ns < 10) {
			printf("Only %d successes out of 28: \n", ns);
			xpcards(Deck, 24);
		}
		Numsucc[ns] += 1;
	}
	for (ix = 0; ix < 29; ix++) {
		printf("Chance of %d successes out of 28 = %f (%d)\n",
			ix, Numsucc[ix] / (double)iter, Numsucc[ix]);
	}
	exit(EXIT_GOOD);
}
