
#include	"poker.h"

#define EXIT_GOOD       0
#define EXIT_PINCON     25      /* arguments inconsistent */
#define EXIT_USAGE      64      /* command line usage error */
#define EXIT_SOFTWARE   70      /* internal software error */

/* This is the famous (but trivial) Fisher-Yates shuffle */
void shuffle(int *Arr, int size)
{
	int	i, j, t;

	for (i = 0; i < size - 1; i++) {
		j = random() % (size - i) + i;
		t = Arr[j];
		Arr[j] = Arr[i];
		Arr[i] = t;
	}
}

/*
 * Routines internal to this module use a
 *    {0,16,32,48} + {2,3,4,5,6,7,8,9,10,11,12,13,14, (15}}
 * format for cards.  But externally we use
 *    {0,1,2,3,4,...,51}
 */

char	name_rank[] =	"--23456789TJQKAj";
char	name_suit[] =	"-SH-D---C------o";
#define	DEUCE		2
#define	JOKER		15

char *name_card(char *out, Card x)
{
	out[0] = name_rank[XRANK(x)];
	out[1] = name_suit[XSUIT(x)];
	out[2] = '\0';
	return out;
}

int s2i(char *inp)
{
	char	*index();
	char	*s, *r;
	char	c1 = inp[0];
	char	c2 = inp[1];

	if ('a' <= c2 && c2 <= 'z')
		c2 += 'A' - 'a';
	if (c2 != 'O' && 'a' <= c1 && c1 <= 'z')
		c1 += 'A' - 'a';
	r = index(name_rank, c1);
	s = index(name_suit, c2);
	if (s && r)
		return (((s - name_suit) << 4) + r - name_rank);
	fprintf(stderr, "Unrecognized card *%s*\n", inp);
	exit(EXIT_PINCON);
}

char	*valnames[] = {
	/* Codes for hand types */
	/* 0 */		"no pair",
	/* 1 */		"1 pair",
	/* 2 */		"2 pair",
	/* 3 */		"triplets",
	/* 4 */		"straight",
	/* 5 */		"flush",
	/* 6 */		"full house",
	/* 7 */		"4-of-a-kind",
	/* 8 */		"str flush",
	/* 9 */		"5-of-a-kind",
};

#if	1
#define	PATHAND(x)	(((x) >> 24) >= 4)
#else
#define	PATHAND(x)	(((x) >> 24) >= 4 && ((x) >> 24) != 7)
#endif

/* Poker hand ranking.
 * Return TBccccc with T type, B tiebreaker (e.g. pair rank)
 *  and ccccc usually the sorted cards (except for 2-pair
 *  where they are needed for tie-breaking on second-pair.)
 */
int pokval(Card c0, Card c1, Card c2, Card c3, Card c4)
{
	register int	ghand;
	register int t, h0, h1, h2, h3, h4;
	int	isflush;
	int	numpair, numtrip;
	int	i;

	/* Detect flush and clear suit bits */
	isflush = c0 & c1 & c2 & c3 & c4 & 0xf0;
	h0 = c0 & 15;
	h1 = c1 & 15;
	h2 = c2 & 15;
	h3 = c3 & 15;
	h4 = c4 & 15;
#define	CSWAP(a, b)	if (a < b) t = a, a = b, b = t
	CSWAP(h0, h1);
	CSWAP(h1, h2);
	CSWAP(h2, h3);
	CSWAP(h3, h4);
	CSWAP(h0, h1);
	CSWAP(h1, h2);
	CSWAP(h2, h3);
	CSWAP(h0, h1);
	CSWAP(h1, h2);
	CSWAP(h0, h1);

	/* Check adjacent cards for pairs */
	numpair = (h0 == h1) + (h1 == h2) + (h2 == h3) + (h3 == h4);
	/* Check near-adjacent cards for triplets */
	numtrip = (h0 == h2) + (h1 == h3) + (h2 == h4);

	if (h0 == JOKER) {
		if (numpair == 0) {
			i = h1 - h4;
			if (i == 3) {
				ghand = (isflush ? 8 : 4) << 24
					| (h1 == DEUCE + 12 ? h1 : h1 + 1);
				goto finished;
			} else if (i == 4) {
				ghand = (isflush ? 8 : 4) << 24 | h1;
				goto finished;
			} else if (h1 == DEUCE + 12 && h2 <= DEUCE + 3) {
				goto fivestr;
			} else if (isflush) {
				h0 = DEUCE + 12;
				if (h1 == DEUCE + 12)
					if (h2 == --h1)
						if (h3 == --h2)
							if (h4 == --h3)
								--h4;
				goto doflush;
			}
		}
		h0 = DEUCE + 12;
		numpair += h1 == DEUCE + 12;
		numtrip += h2 == DEUCE + 12;
	} else if (numpair == 0) {
		i = h0 - h4;
		if (i == 4) {
			ghand = (isflush ? 8 : 4) << 24 | h0;
			goto finished;
		} else if (i == 12 && h1 == DEUCE + 3) {
		fivestr:
			ghand = (isflush ? 8 : 4) << 24 | (DEUCE + 3);
			goto finished;
		} else if (isflush) {
		doflush:
			ghand = (5) << 24;
			ghand |= h0 << 16 | h1 << 12 | h2 << 8 | h3 << 4 | h4;
			goto finished;
		}
	}

	/* Switch on hand type */
	switch (numpair + numtrip) {
	case 0:
	/* No pair */
		ghand = (0) << 24;
		ghand |= h0 << 16 | h1 << 12 | h2 << 8 | h3 << 4 | h4;
		break;
	case 1:
	/* One pair */
		ghand = (1) << 24;
		ghand |= h0 << 16 | h1 << 12 | h2 << 8 | h3 << 4 | h4;
		/* Forgot where the pair is; find it again for tiebreaker */
		if (h0 == h1 || h1 == h2)
			ghand |= h1 << 20;
		else
			ghand |= h3 << 20;
		break;
	case 2:
	/* Two pair */
		ghand = (2) << 24;
		ghand |= h1 << 16 | h3 << 12 | h0 << 8 | h2 << 4 | h4;
		break;
	case 3:
	/* Triplets */
		ghand = (3) << 24;
	tiebreak:
		ghand |= h2 << 20 | h0 << 16 | h1 << 12
				| h2 << 8 | h3 << 4 | h4;
		break;
	case 4:
	/* Full house */
		ghand = (6) << 24;
		goto tiebreak;
	case 5:
	/* Four-of-kind */
		ghand = (7) << 24;
		goto tiebreak;
	case 7:
	/* Five-of-kind */
		ghand = (9) << 24;
		goto tiebreak;
	default:
		printf("numpair = %d numtrip = %d\n", numpair, numtrip);
		exit(EXIT_SOFTWARE);
	}
	finished:
	return ghand;
}

void pcards(Card *han, int cnt)
{
	int	i;
 
	for (i = 0; i < cnt; i++) {
		char    cardname[3];
		name_card(cardname, han[i]);
		printf("%s ", cardname);
	}
}

void xpcards(Card *han, int cnt)
{
	int	i;
	char	c;
 
	for (i = 0; i < cnt; i++) {
		char    cardname[3];
		c = CX2I(han[i]);
		name_card(cardname, han[i]);
		printf("%s ", cardname);
	}
}

int phand(char *msg, Card *han)
{
	int	val;

	pcards(han, 5);
	val = pokval(han[0], han[1], han[2], han[3], han[4]);
	printf(" ( %08x %s ) -- %s\n", val, valnames[val >> 24], msg);
	return val >> 24;
}

