
/*
 * Help generate word arithmetic puzzles.
 *	wordarith SEND + MORE > temporary
 *	grep -v = < temporary | sort | uniq -u > Wordarith.sols
 * Inspect Wordarith.sols and see if there are any words you like.
 * In this example you will find NODES and MONE,
 * The comma in  MONE,  invites you to pick the letter.
 *
 * NB: This program is NOT optimized for speed (it's rather slow)
 * Nor for elegance, readability, etc.  Sorry.
 */
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>

#define	NBASE	10

char	BaseLets[30];
char	Lets[]	= ".........................";
int	Val[256];
int	Verbose = 0;

void setletts(char *s)
{
	int	c, i = 0;

	while (c = *s++) {
		if (! strchr(BaseLets, c)) {
			BaseLets[i++] = c;
		}
	}
	if (i < 5 || i > 10) {
		printf(
"Expecting 5 to 10 distinct letters.  Instead there were %d\n", i);
		exit(1);
	}
	c = '1';
	while (i < NBASE) {
		BaseLets[i++] = c++;
	}
}

int redund(void)
{
	int	i, gots;

	for (gots = i = 0; i < NBASE; i++) {
		switch (Lets[i]) {
		case '1':
			if (++gots != 1)
				return 1;
			break;
		case '2':
			if (++gots != 2)
				return 1;
			break;
		case '3':
			if (++gots != 3)
				return 1;
			break;
		case '4':
			if (++gots != 4)
				return 1;
			break;
		case '5':
			if (++gots != 5)
				return 1;
			break;
		default:
			break;
		}
	}
	return 0;
}

void fixy(char *s)
{
	char	mappy[26], c;
	int	nummapp = 0, i;

	for (i = 0; i < 26; i++)
		mappy[i] = 0;
	for (i = 0; c = s[i]; i++) {
		if ('1' <= c && c <= '5') {
			if (mappy[c - '1'] == 0) {
				mappy[c - '1'] = 'a' + nummapp++;
			}
			s[i] = mappy[c - '1'];
		}
	}
}

long long s2v(char *s)
{
	long long	res = 0;
	char	c;

	if (Val[*s] == 0)
		return -1;
	while (c = *s++) {
		res *= NBASE;
		res += Val[c];
	}
	return res;
}

char *v2s(long long v)
{
	int	i;
	static char res[20];

	for (i = 15; v; i--) {
		res[i] = Lets[v % NBASE];
		v /= NBASE;
	}
	return res + i + 1;
}

void setperm(long long x)
{
	int	k, m, t;

	for (k = 0; k < NBASE; k++)
		Lets[k] = BaseLets[k];
	for (k = NBASE; k > 1; k--) {
		m = x % k;
		t = Lets[k-1];
		Lets[k-1] = Lets[m];
		Lets[m] = t;
		x /= k;
	}
	for (k = 0; k < NBASE; k++)
		Val[Lets[k]] = k;
}

void upperize (char *s)
{
	int	c;

	while (c = *s) {
		if ('a' <= c && c <= 'z')
			*s += 'A' - 'a';
		++s;
	}
}

int main(int argc, char **argv)
{
	long long	t, x, v1, v2, v3;
	long long	maxx;
	char	*s, *w1, *w2, *w3, op, op2;
	char	all_lets[30];

	setlinebuf(stdout);
	Lets[NBASE] = 0;

	w1 = argv[1];
	upperize(w1);
	op = argv[2][0];
	w2 = argv[3];
	upperize(w2);
	strcpy(all_lets, w1);
	strcat(all_lets, w2);
	if (argc == 6) {
		op2 = argv[4][0];
		w3 = argv[5];
		upperize(w3);
		strcat(all_lets, w3);
	} else {
		w3 = 0;
		op2 = 0;
	}
	setletts(all_lets);

	for (maxx = x = 1; x <= NBASE; x++)
		maxx *= x;
	for (x = 0; x < maxx; x++) {
		setperm(x);
		if (redund()) {
			if (Verbose)
				printf("%s Redundant =\n", Lets);
			continue;
		}
		v1 = s2v(w1);
		v2 = s2v(w2);
		v3 = argc == 6 ? s2v(w3) : 0;
		if (v1 < 0 || v2 < 0 || v3 < 0) {
			if (Verbose)
				printf("%s Leading zero =\n", Lets);
			continue;
		}
		switch (op) {
		case '+': t = v1 + v2; break;
		case '-': t = v1 - v2; break;
		case 'X': case '*':
		case 'x': t = v1 * v2; break;
		case '/': t = v1 / v2;
			if (v2 * t != v1)
				goto notdiv;
			break;
		default:
			printf("Bad op\n");
			exit(1);
		}
		if (argc == 6)
		switch (op2) {
		case '+': t += v3; break;
		case '-': t -= v3; break;
		case 'r': t = v3 - t; break;
		case 'X': case '*':
		case 'x': t *= v3; break;
		case '/': if (t % v3)
				goto notdiv;
			t = t / v3;
			break;
		default:
			printf("Bad op\n");
			exit(1);
		}
		if (t < 0) {
			if (Verbose)
				printf("%s Subtract fails =\n", Lets);
		notdiv:
			continue;
		}
		s = v2s(t);
		fixy(s);
		printf("%s\n", s);
		if (argc == 6)
			printf("%s (%s %c %s) %c %s == %s\n",
				Lets, w1, op, w2, op2, w3, s);
		else
			printf("%s %s %c %s == %s\n",
				Lets, w1, op, w2, s);
	}

	exit(0);
}
