
#include	<stdio.h>
#include	<stdlib.h>
#include	<math.h>
#include	<string.h>


#define EXIT_GOOD       0
#define EXIT_SFAIL      20      /* search failed */
#define EXIT_USAGE      64      /* command line usage error */

int	Numshow = 12;
double	Minmagshow = 0.1;

/* 5 log(d/10) */

#define	calcdist(a, b, c, d, e, f)	\
	sqrt(((a)-(b))*((a)-(b)) + ((c)-(d))*((c)-(d)) + ((e)-(f))*((e)-(f)))

struct sstar {
	char	*nam1, *nam2;
	char	*sdist, *smag, *samag;
	char 	*sspect;
	char 	*sx, *sy, *sz;
} SStar[] = {
#ifdef	BIGLIST
#define	BIGLIST	/* Defaults to off */
#include	"big_starlist.h"
#else
#include	"starlist.h"
#endif
};

#define	NUMEL(arr)	(sizeof (arr) / sizeof *(arr))

struct star {
	char 	*name;
	double	dist, ndis, mag, amag, x, y, z;
	double	rmag;
} Star[260000];
int	Numstar;

struct star *findstar(char *nam)
{
	struct star *p;

	for (p = Star; p < Star + Numstar; p++)
		if (! strcmp(p->name, nam))
			return p;
	return 0;
}

double	Tco[3], Oco[3], A[3], Dir[3];
double	co_del[3], co_mid[3];

int coco(const void *p1, const void *p2)
{
	double	f1 = ((struct star *)p1)->rmag;
	double	f2 = ((struct star *)p2)->rmag;

	if (f1 < f2)
		return -1;
	else if (f1 > f2)
		return 1;
	else
		return 0;
}

/* Given amag and distance, return mag */
double	mkmag(double amag, double dist)
{
	double	res;

	res = exp(- amag * 0.92103403719762) / (dist * dist / 100);
	res = -log(res) / 0.92103403719762;
	return res;
}

int desired (char *s)
{
	return 0
		|| !strcmp(s, "29____Ari")
		|| !strcmp(s, "17Iot_Psc")
		|| !strcmp(s, "6Pi__Sco")
		|| !strcmp(s, "22____Sco")
		|| !strcmp(s, "5Rho_Oph")
		|| !strcmp(s, "hr6193")
		|| !strcmp(s, "hr152429")
		|| !strcmp(s, "hr6630")
		|| !strcmp(s, "Achernar")
		|| !strcmp(s, "Eta_Sgr")
		|| !strcmp(s, "Sheratan")
		|| !strcmp(s, "Aldebaran")
		|| !strcmp(s, "17Eta_Ari")
		|| !strcmp(s, "47____Ari")
		|| !strcmp(s, "41____Ari")
		|| !strcmp(s, "58Zet_Ari")
		|| !strcmp(s, "Taygete");
}

void dosurvey(double lyyet)
{
	int	i, besti;
	double	dis, bestdis = 100000, nx, ny, nz, resmag;
	double	bestmag = 999, best2mag = 999, closemag = 0;
	char	*bestn, *bestn2, *closen = 0;
	struct star *p;

	printf("\n");
	for (p = Star, i = 0; i < Numstar; i++, p++) {
		nx = p->x - Tco[0];
		ny = p->y - Tco[1];
		nz = p->z - Tco[2];
		p->ndis = dis = sqrt(nx*nx + ny*ny + nz*nz);
		p->rmag = mkmag(p->amag, dis);
		if (0) if (!strcmp(p->name, "Kaus_Australis"))
			printf("%s %4.2f KLUGE at %7.2f ly\n",
				p->name, p->rmag, lyyet);
		if (p->rmag < -1 || i < 3 || p->rmag < 0 && desired(p->name))
			printf("%s %4.2f BRIGHT at %7.2f ly\n",
				p->name, p->rmag, lyyet);
		if (p->rmag < best2mag) {
			if (p->rmag < bestmag) {
				best2mag = bestmag;
				bestn2 = bestn;
				bestmag = p->rmag;
				bestn = p->name;
			} else {
				best2mag = p->rmag;
				bestn2 = p->name;
			}
		}
		if (dis < bestdis) {
			closen = p->name;
			bestdis = dis;
			closemag = p->rmag;
		}
	}
	printf(" %6.3f Closest star = %s [%5.2f, Mag=%3.2f] ",
			lyyet,
			closen, bestdis * 3.26156, closemag);
	qsort(Star, Numstar, sizeof *Star, coco);
	printf("\n RK: ");
	for (i = 0; i < Numshow; i++) {
		if (i > 3 && Star[i].rmag > Minmagshow)
			break;
		printf("%s [%4.2f] ", Star[i].name, Star[i].rmag);
	}
	for (; Star[i].rmag < 0; i++) {
		if (!strcmp("Rigel", Star[i].name))
			printf("\n Rigel Rigel Rigel xxxx\n");
		if (desired(Star[i].name))
			printf("%s [%4.2f] ",
				Star[i].name, Star[i].rmag);
	}

	printf("\n");
}

double calc_dist(double x1, double x2, double x3,
		double y1, double y2, double y3) 
{
	return sqrt(0
			+ (x1 - y1) * (x1 - y1)
			+ (x2 - y2) * (x2 - y2)
			+ (x3 - y3) * (x3 - y3));
}

double	Radius;

int	nice(double x)
{
	int	i = x;
	if (x - i > 0.1)
		return 0;
	return i % 10 == 0;
}

double	LY_PER = 1;

int main(int argc, char **argv)
{
	struct sstar *sp;
	struct star *p, *ptarg1, *ptarg2;
	int	iter, tgt;
	unsigned int i;
	double	magg, ff, itt;
	double	Edist, dis;
	double	nd, Bnd1 = 0, Bnd2 = 0;
	int	get_coo_beenhere = 0;

	argloop:
	if (argc < 3) {
		printf("Usage: starbright [options] starname_1 starname_2\n");
		exit(EXIT_USAGE);
	}
	if (!strcmp(argv[1], "-minmag")) {
		Minmagshow = atof(argv[2]);
		argc -= 2;
		argv += 2;
	}
	if (!strcmp(argv[1], "-maxshow")) {
		Numshow = atof(argv[2]);
		argc -= 2;
		argv += 2;
		goto argloop;
	}
	if (!strcmp(argv[1], "-bnd")) {
		Bnd1 = atof(argv[2]);
		Bnd2 = atof(argv[3]);
		argc -= 3;
		argv += 3;
		goto argloop;
	}
	if (argc > 3)
		LY_PER = atof(argv[3]);
	setlinebuf(stdout);
	for (sp = SStar, p = Star, i = 0; i < NUMEL(SStar); i++, sp++) {
		p->amag = atof(sp->samag);
		if (p->amag > 11)
			continue;
		p->dist = atof(sp->sdist);
		p->rmag = p->mag = atof(sp->smag);
		p->x = atof(sp->sx);
		p->y = atof(sp->sy);
		p->z = atof(sp->sz);
		if (!strcmp(sp->nam2, "XX")) {
			if (!strcmp(sp->nam1, "XX")) {
				p->name = malloc(50);
				p->name[0] = '#';
				strncpy(p->name + 1, sp->sx, 48);
				if (0) printf("Missing name at line %d\n", i);
			} else {
				p->name = sp->nam1;
			}
		} else {
			p->name = sp->nam2;
		}
		if (p->dist > 90000) {
			p->dist /= 100;
			p->x /= 100;
			p->y /= 100;
			p->z /= 100;
			p->amag += 10.000;

			p->dist /= 3;
			p->x /= 3;
			p->y /= 3;
			p->z /= 3;
			p->amag += 2.3856;

			p[1] = p[0];

			p++;
			p->name += 1;	/* ! */

			p->dist *= 10;
			p->x *= 10;
			p->y *= 10;
			p->z *= 10;
			p->amag -= 5.000;

			p->dist *= 2;	/* resp 4, 5 */
			p->x *= 2;
			p->y *= 2;
			p->z *= 2;
			p->amag -= 1.505;	/* resp 3.0103, 3.4948 */
		}
		p++;
	}
	Numstar = p - Star;

	if (0)
		printf("Numstar = %d\n", Numstar);
	ptarg1 = findstar(argv[1]);
	ptarg2 = findstar(argv[2]);
	if (ptarg1 == 0 || ptarg2 == 0) {
		printf("Target not found %s %s\n",
				argv[1], argv[2]);
		exit(EXIT_SFAIL);
	}

	Dir[0] = ptarg1->x - ptarg2->x;
	Dir[1] = ptarg1->y - ptarg2->y;
	Dir[2] = ptarg1->z - ptarg2->z;
	if (0) calcdist(Dir[0], Dir[1], Dir[2], Dir[0], Dir[1], Dir[2]);
	dis = sqrt(Dir[0] * Dir[0] + Dir[1] * Dir[1] + Dir[2] * Dir[2]);
	printf("Total distance %s to %s is %f parsecs (%f l-y)\n",
		argv[1], argv[2], dis, 3.26156 * dis);

	dis *= 3.26156 / 100;	/* ly in centi-trip */
	itt = 0;
	while (itt > - 20)
		itt -= LY_PER / dis;

	Oco[0] = Tco[0] = ptarg1->x;
	Oco[1] = Tco[1] = ptarg1->y;
	Oco[2] = Tco[2] = ptarg1->z;
	for (; itt <= 120.1; itt += LY_PER / dis) {
		if (itt < Bnd1 || itt > Bnd2)
			continue;
		if (0) printf("%f %f %f --- ", Tco[0], Tco[1], Tco[2]);
		Tco[0] = Oco[0] - Dir[0] * (itt / 100.0);
		Tco[1] = Oco[1] - Dir[1] * (itt / 100.0);
		Tco[2] = Oco[2] - Dir[2] * (itt / 100.0);
		if (0) printf("%f %f %f --- ", Tco[0], Tco[1], Tco[2]);
		if (0) printf("%f %f %f\n",
					Dir[0] / 100.0,
					Dir[1] / 100.0,
					Dir[2] / 100.0);
		Edist = calc_dist(Tco[0], Tco[1], Tco[2],
				0.0, 0.0, 0.0);
		nd = 3.26156 * calc_dist(Tco[0], Tco[1], Tco[2],
				Oco[0], Oco[1], Oco[2]);
		if (itt < 0)
			nd = -nd;
		printf("%7.2f ly along path (Earth %f) --> ",
				nd, 3.26156 * Edist);
		dosurvey(nd);
	}

	exit(EXIT_GOOD);
}
