
/* Copyright (c) 2008 James Dow Allen.  All Rights Reserved; see License.
 * Macros and routines for general bit store/fetch
 */

/* One *might* want 64-bit Bgroups, but Vdata 32-bit.
 * In fact, for our application setkey() might have small
 *  v, setslot() long.  But we don't support any of that
 * now.
 * Another frill that might boost performance is to replace
 * our basic data
 *		void foo(Uint64 *p, int o, int k);
 *		Uint64	b_ddd[1000];
 *		foo(b_ddd[bit_off / 64], off % 64, k);
 * with
 *		Uint32  b_ddh[2000];
 *		foo((Uint64 *)(b_ddh[bit_off / 32]), 32, k);
 */
typedef Bgroup Vdata;

/* Scramble/unscramble macro */
#define	SCRAM(x, ig, mpy, p)				\
	(((x) >> (ig)) * (Uint64)(mpy) % (p) << (ig)	\
		| (x) & (One64 << (ig)) - 1)

/* We assume a typedef for Bgroup has already been provided */
#define	BITS_BG		(8 * sizeof(Bgroup))
#define	ALLOSZ(x)	(((x) + BITS_BG - 1) / BITS_BG)
#define	B_One	((Bgroup)1)

#define	GETBITS(p, k, n)				\
	((*(p) >> BITS_BG - (k) - (n)) & (B_One << (k)) - 1)

#define	SJS0(p, k, v)					\
	(*(p) = *(p) & ((B_One << BITS_BG -(k)) - 1)	\
		| (v) << BITS_BG - (k))

#define	SJS1(p, k, n, v)				\
	(*(p) = *(p) & (~((B_One << BITS_BG - (n))	\
		  - (B_One << BITS_BG - (n) - (k))))	\
		| (v) << BITS_BG - (n) - (k))

#define	SETBITS(p, k, n, v)				\
	((n) ? SJS1(p, k, n, v) : SJS0(p, k, v))

static void setb_gen(Bgroup *p, Uint k, Uint n, Vdata v)
{
	Uint kn;

	p += n / BITS_BG;
	n %= BITS_BG;
	if ((kn = k + n) > BITS_BG) {
		SJS1(p, BITS_BG - n, n, v >> kn - BITS_BG);
		SJS0(p + 1, kn - BITS_BG, v);
	} else {
		SETBITS(p, k, n, v);
	}
}

static Vdata getb_gen(Bgroup *p, Uint k, Uint n)
{
	Uint kn = k + n % BITS_BG;

	p += n / BITS_BG;
	return (kn > BITS_BG
			? p[0] << kn - BITS_BG | p[1] >> BITS_BG + BITS_BG - kn
			: p[0] >> BITS_BG - kn)
		& (B_One << k) - 1;
}

#define	Jsetslot(p, n, v, e)			\
	setb_gen((p), (e), (n)*(e), (v))

#define	Jsetkey(p, n, v, s, e)			\
	setb_gen((p), (s), (n)*(e), (v))

#define	Jsetpay(p, n, v, s, e)			\
	((s) ? setb_gen((p), (s), (n)*(e) + (e) - (s), (v)) : 0)

#define	Jgetslot(p, n, e)			\
	getb_gen((p), (e), (n)*(e))

#define	Jgetkey(p, n, s, e)			\
	getb_gen((p), (s), (n)*(e))

#define	Jgetpay(p, n, s, e)			\
	((s) ? getb_gen((p), (e), (n)*(e) + (e) - (s)) : 0)

