
/* Copyright (c) 2008 James Dow Allen.  All Rights Reserved; see License. */
/* Here follows hash routines' interface to the world */

/* Applications need know nothing about struct htab_tab */
struct htab_tab;

typedef enum {
	HLook = 1, HArep = 2, HHrep = 3, HLrep = 4, Hadd = 5,
	HDelete = 6,
	HCreate = 7,
	HAmak = 8, HHmak = 9, HLmak = 10, HCadd = 11,
} htab_op;

struct htab_trav {
	long	wkreset;
	long	wkindx, wkoix;
	short	wktab, wkslot;
};

struct htab_params {
	Key	maxkey;
	int	qbits, pbits, mincuck, maxcuck;
};

/* Since payload is never negative, 1's complement is
 * a convenient way to flag "old entry".  This is chosen
 * to minimize what caller needs to know of table internals.
 *
 * (-1) is used for error return, but ...
 * ... errors "won't happen"!     :-)   :-)
 */
#define	htab_OLDE(x)		(~(x))
#define	htab_AENTRY(x)		(~(x))

/*
 * Find g in the Table, possibly updating its payload
 * to be val.
 * Return htab_OLDE(payload) if old entry exists
 * 0 if not found (or payload for new insertion)
 */
int htab_find(struct htab_tab *tabp, Key g, int val, htab_op hop, int *tagp);
/* Create |numtabs| tables, setting them to empty */
void htab_create(struct htab_tab **p, int numtabs, struct htab_params *pp);
/* Return (in *gp) next entry in a walk through Hash Table,
 * initiating a new search of resetf is non-zero.
 * Return payload of found entry, or 0 at end of search.
 */
int htab_getnxt(struct htab_tab *tabp, struct htab_trav *wp, Key *gp);
void htab_sstat(struct htab_tab *tabp);
Uint64 htab_maxkey(struct htab_tab *tabp);
Uint64 htab_maxtag(struct htab_tab *tabp);
Uint htab_maxpoint(struct htab_tab *tabp);

