
/* Copyright (c) 2008 James Dow Allen.  All Rights Reserved; see License. */
/* Prepare parameters for tight hash table */
#include	"js_simple.h"

union {
	unsigned long b[1];
	struct {
		unsigned long a0:8, a1:8, a2:8, a3:8;
	} s[1];
} u[4];

union {
	unsigned long long b[1];
	struct {
		unsigned long a0:8, a1:8, a2:8, a3:8;
		unsigned long b0:8, b1:8, b2:8, b3:8;
	} sl[1];
} ul[8];

int locbf(int numby)
{
	int	i;

	u[0].s[0].a0 = 0xff;
	u[1].s[0].a1 = 0xff;
	u[2].s[0].a2 = 0xff;
	u[3].s[0].a3 = 0xff;
	ul[0].sl[0].a0 = 0xff;
	ul[1].sl[0].a1 = 0xff;
	ul[2].sl[0].a2 = 0xff;
	ul[3].sl[0].a3 = 0xff;
	ul[4].sl[0].b0 = 0xff;
	ul[5].sl[0].b1 = 0xff;
	ul[6].sl[0].b2 = 0xff;
	ul[7].sl[0].b3 = 0xff;
	for (i = 0; i < numby; i++)
		if (0xff == (numby == 4
				? u[i].b[0]
				: ul[i].b[0] >> 32))
			return i;
	fprintf(stderr, "Bitfield location failed\n");
	exit(EXIT_SOFTWARE);
}

void showpad(int npad)
{
	printf("\tunsigned long ");
	if (npad > 0) printf("pad0:8, ");
	if (npad > 1) printf("pad1:8, ");
	if (npad > 2) printf("pad2:8, ");
	printf("tmmai:2, tmofa:3, tmofb:3");
	if (npad < 3) printf(", pad3:8");
	if (npad < 2) printf(", pad4:8");
	if (npad < 1) printf(", pad5:8");
	printf(";\n");
}

int main(int argc, char **argv)
{
	long	q, tsize, a, memreq, omemreq;
	unsigned long res;
	double	megneed, exptab;
	long long n;
	int	j, igbits = 0, bsize;
	int	qq, slotper, minsiz, paysize;
	int	npad;
	int	skipquo;
	int	spp2 = 0;

	if (argc != 6) {
		fprintf(stderr, "\nUsage to set parameters for Tight Hash:\n");
		fprintf(stderr, "\tgenparms #N #Q #P #Blk #Minsize\n");
		fprintf(stderr, " where N is max key, Q is numb bits for quotient,\n");
		fprintf(stderr, " P is numb payload bits, Blk is slots per bucket,\n");
		fprintf(stderr, " and Minsize is minimum total number of slots.\n");
		fprintf(stderr, "\nPLEASE NOTE: genparms doesn't optimize all parameters itself:\n");
		fprintf(stderr, "  Try to tune the arguments to genparms.\n");
		exit(EXIT_USAGE);
	}
	n = atoll(argv[1]);
	qq = atoi(argv[2]);
	paysize = atoi(argv[3]);
	slotper = atoi(argv[4]);
	minsiz = atoi(argv[5]);
	if (paysize == 0)
		skipquo = 2;	/* leave room for "deleted" */
	else
		skipquo = 1;
	if (slotper < 12) {
		fprintf(stderr, "Tight-Hash bucket size must be at least 12\n");
		fprintf(stderr, "(Between 15 and 65 is good ... lower for speed, higher for space.)\n");
		slotper = 12;
	} else if (slotper > 90) {
		fprintf(stderr, "Large bucket size is slow.\n");
		fprintf(stderr, "(Between 15 and 65 is good ... lower for speed, higher for space.)\n");
		fprintf(stderr, "Instead perhaps decrease #Q.\n");
		fprintf(stderr, "(This is a non-fatal warning.)\n");
	}

	if (n < (long long)1 << 32)
		printf("typedef unsigned long Key;\n");
	else
		printf("typedef unsigned long long Key;\n");
	q = (1ul << qq) - skipquo;
	/* We now shrink n to fit nicely in Uint32 just for rest routine */
	for (igbits = 0; n > 1 << 30; n >>= 1, igbits++)
		;
	printf("#define IGBITS\t%d\n", igbits);
	printf("#define QSIZE\t%d\n", qq);
	printf("#define PSIZE\t%d\n", paysize);
	printf("#define SKIPQUO\t%d\n", skipquo);
	if (qq + paysize < 32) {
		printf("typedef unsigned long Bgroup;\n");
		bsize = 32;
		npad = locbf(4);
		printf("struct bg2bf {\n");
		showpad(npad);
		printf("};\n");
	} else {
		printf("typedef unsigned long long Bgroup;\n");
		bsize = 64;
		npad = locbf(8);
		printf("struct bg2bf {\n");
		if (npad > 3) {
			printf("\tunsigned long padw;\n");
			showpad(npad - 4);
		} else {
			showpad(npad);
			printf("\tunsigned long padw;\n");
		}
		printf("};\n");
	}

	tsize = ((n << igbits) / q) * slotper;
	exptab = (minsiz / (double)tsize) * 1.1;
	if (exptab < 0.7) {
		fprintf(stderr,
			"Note: Table bigger than target (%f) -- perhaps tune parms\n",
			exptab);
	}
	if (exptab > 2.0) {
		fprintf(stderr,
			"#Q set bigger than necessary (%f) -- should tune parms\n",
			exptab);
	}
	if (exptab > 1)
		n *= exptab;
	if (n != (n & 0xffffffff)) {
		fprintf(stderr, "Big n (%lld)\n", n);
		exit(EXIT_EXCEED);
	}
	for (; ! isprime(n); n += 1)
		;
	tsize = (n << igbits) / q;
	for (tsize++; ! isprime(tsize); tsize++)
		;
	printf("#define TABSZ01 %ld\n", tsize);
	if (tsize * 2 + 1 < One64 << qq + paysize) {
		printf("#define SLOTPP\t%d\n", 1);
		printf("#define HALFADR\t%d\n", 0);
	} else if (tsize + 2 < One64 << qq + paysize) {
		printf("#define SLOTPP\t%d\n", 1);
		printf("#define HALFADR\t%d\n", 1);
	} else if ((tsize * 2 >> qq + paysize) + 1 < One64 << qq + paysize) {
		spp2 = 1;
		printf("#define SLOTPP\t%d\n", 2);
		printf("#define HALFADR\t%d\n", 0);
	} else {
		fprintf(stderr, "Slots too small for pointers.\n");
		fprintf(stderr, "Increase #Q or redesign tightch.c.\n");
		exit(EXIT_PINCON);
	}
	printf("#define TOTSIZE\t%d\n", slotper);
	/* Simulations show following A/B sizes to be efficient */
	printf("#define ABSIZE\t%d\n",
			slotper > 50 ? 5 :
			(spp2 || slotper > 20) ? 4 : 3);
	printf("#define BBSIZE\t%d\n",
			(spp2 || slotper > 70) ? 4 : 3);
	omemreq = memreq = (paysize + qq) * slotper + 8;
	memreq = ((memreq + bsize - 1) / bsize) * (bsize / 8);
	printf("/* Bucket utilization = %f */\n",
			omemreq / (memreq * 8.0));
	megneed = memreq * (tsize / (1024 * 1024.0));
	a = n * .61803398875 - 2;
	while ((res = mk_revh(a, n)) == 0)
		++a;
	printf("#define REV_PRIM\t%lu\n", n);
	printf("#define REV_FMPY\t%lu\n", a);
	printf("#define REV_IMPY\t%lu\n", res);
	printf("/* %f Megabytes for Hash Table */\n", megneed);
	tsize *= slotper;
	tsize *= spp2 ? .86 : .88;
	printf("/* Good for about %d entries */\n", tsize);
	exit(EXIT_GOOD);
}

