
#include	"js_simple.h"

/* Copyright (c) 2008 James Dow Allen.  All Rights Reserved; see License.
 * Subroutines useful for generating
 * reversible hash functions.
 */

int isprime(Uint32 x)
{
	Uint32 y, z;

	if (x % 2 == 0)
		return 0;
	for (y = 3; z = x / y, z >= y; y += 2)
		if (x == z * y)
			return 0;
	return 1;
}

/* Find the first prime larger than x */
Uint32 nxtprime(Uint32 x)
{
	Uint	xx = x % 6;
	Uint32	tx = x - xx + 1;

	if (x < 7)
		return x < 3 ? 3 : x < 5 ? 5 : 7;
	/* Next is a bit like Duff's famous Device */
	else switch (xx) {
	case 0:
		while (1) {
			if (isprime(tx))
				return tx;
		/* FALLTHRU */
	case 1: case 2: case 3: case 4:
			if (isprime(tx + 4))
				return tx + 4;
		/* FALLTHRU */
	case 5:
			tx += 6;
		}
	}
	/* NOTREACHED */
	return 0; /* but suppress warning message anyway */
}

/* Compute a^b (mod n) */
static unsigned long expmod(unsigned long a, unsigned long b, unsigned long n)
{
	long ix, resu;

	for (ix = 1L << 30; b && !(ix & b); ix >>= 1)
		;
	resu = a;
	for (ix >>= 1; ix; ix >>= 1) {
		resu = (resu * (long long)resu) % n;
		if (ix & b)
			resu = (resu * (long long)a) % n;
	}
	return resu;
}

static unsigned long gfactor(unsigned long x)
{
	unsigned long q;

	if (!(x & 1))
		return 2;
	for (q = 3; q * q < x; q += 2)
	if (x % q == 0)
		return q;
	return x;
}

unsigned long mk_revh(unsigned long a, unsigned long p)
{
	unsigned long res, qq, t, ot;

	qq = p - 1;
	for (ot = -1; (t = gfactor(qq)) > 1; ot = t) {
		qq /= t;
		if (t == ot)
			continue;
		res = expmod(a, (p - 1) / t, p);
		if (res < 2)
			return 0;
	}
	res = expmod(a, p - 2, p);
	return res;
}
