
/* Copyright (c) 2020 James Dow Allen.  All Rights Reserved; see License. */

#define		MAIN_MODULE
#include	"js_simple.h"

/*
 * Develop a reversible scrambler,
 *	where n is max input,
 */
static void revgen(unsigned long n)
{
	Uint32	newn, a, res;
	int	igbits = 0;

	/*
	 * Remember to check for bizarre params that
	 * would throw us for a loop.
	 */
	if (n > 255 << 22) {
		printf("Sorry: Argument too big\n");
		return;
	}
	newn = n;
	newn = nxtprime(newn - 1);
	a = newn * .61803398875 - 2;
	while ((res = mk_revh(a, newn)) == 0)
		++a;
	printf("Hashcode: (x * %u %% %u)\n", a, newn);
	printf("Inverse:  (x * %u %% %u)\n", res, newn);
}

int main(int argc, char **argv)
{

	if (argc != 2) {
		printf("Usage: reversible #N\n");
		printf("Develop a reversible hash for numbers n = 0, 1, ..., N\n");
	} else {
		revgen(atoll(argv[1]));
	}
	exit(0);
}

