
/* Copyright (c) 2009 James Dow Allen.  All Rights Reserved; see License.
 * These are general definitions for wide variety of software.
 *
 * We use Bigint for the compiler's largest unsigned integer type;
 * Yes I know these first lines will thus have to be altered
 *   to port to some systems.
 */

#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	<limits.h>

// typedef	unsigned int size_t;
typedef	unsigned int Uint;
typedef	unsigned long Uint32;
typedef	unsigned long long Uint64;
#define	One32	((Uint32)1)
#define	One64	((Uint64)1)

typedef unsigned long long Bigint;
typedef long long SBigint;
#define BITS_IN_BIGINT  (64)

#define	GCD_Routine(Name, Tyype)		\
Tyype Name(Tyype a, Tyype b)			\
{						\
						\
	if (a == 0 || b == 0)			\
		return 0;			\
	while (1)				\
		if (a <= b && (b %= a) < 2)	\
			return b ? 1 : a;	\
		else if ((a %= b) < 2)		\
			return a ? 1 : b;	\
}

/* Divide (positive) rounding up */
#define	DIV_RUP(x, y)	(((x) + (y) - 1) / (y))

/* Number of bits in m, where m= (1<<b)-1 */
#define	IMAX_BITS(m)	\
	 ((int)((m) /((m)%255+1)/255%255*8 + 7-100/((m)%255+14)))

#define	BITS_IN_UINT	IMAX_BITS(UINT_MAX)
#define	BITS_IN_ULONG	IMAX_BITS(ULONG_MAX)

#define	PASTE1(a, b)	a##b
#define	PASTE(a, b)	PASTE1(a, b)

/* Exit codes */
#define	EXIT_GOOD	0
#define	EXIT_EXPECT	4	/* "intended" error */
#define	EXIT_BADFILE	6	/* missing file, etc. */
#define	EXIT_MEMORY	16	/* memory could not be obtained */
#define	EXIT_EXCEED	17	/* method limit reached */
#define	EXIT_PEXCESS	18	/* method parameter exceeded */
#define	EXIT_SFAIL	20	/* search failed */
#define	EXIT_SWCONF	24	/* software misconfigured */
#define	EXIT_PINCON	25	/* arguments inconsistent */
#define	EXIT_INVOKE	26	/* incorrect method invocation */

#define	EXIT_USAGE	64	/* command line usage error */
#define	EXIT_SOFTWARE	70	/* internal software error */

/* Macros for a trivial PRNG; Two different trivial rngs are offered here. */
#if	0

#ifndef	MAIN_MODULE
extern
#endif
unsigned long long Simp_seed;
#define	SIMP_RANGE	3787109
#define	SIMP_RNG	(++Simp_seed * 2340563 % SIMP_RANGE)
#define	SIMP_RPROB	((SIMP_RNG * 2 + 1) / (2.0 * SIMP_RANGE))

#else

#ifdef	MAIN_MODULE
unsigned long Simp_seed = 123456789;
#else
extern unsigned long Simp_seed;
#endif
#define	SIMP_RNG	(Simp_seed = Simp_seed * 1103515245 + 12345)
#define	SIMP_RPROB	((SIMP_RNG & 0xffffff | 1) / (1.0 * 0x1000000))

#endif

#define	SIMP_SEED(x)	(Simp_seed = (x))
#define	SIMP_RIX(siz)	(SIMP_RNG % (siz))

/* entry points in simplesub.c */
void *mustmalloc(size_t siz, char *msg);
unsigned long js_gcd(unsigned long a, unsigned long b);
Bigint js_gcdbig(Bigint a, Bigint b);
int js_bitcnt(unsigned long g);
int js_bbitcnt(Bigint g);
int locbit(int n, unsigned long m);
int loclbit(int n, unsigned long m);
/* entry points in pharith.c */
unsigned long mk_revh(unsigned long a, unsigned long p);
unsigned long nxtprime(unsigned long x);
int isprime(unsigned long x);


/* Hand-rolled bitfields */
void js_setbits(unsigned long *p, int bcnt, unsigned long boff,
		unsigned long datum);
unsigned long js_getbits(unsigned long *p, int bcnt, unsigned long boff);
void js_bsetbits(Bigint *p, int bcnt, unsigned long boff, Bigint datum);
Bigint js_bgetbits(Bigint *p, int bcnt, unsigned long boff);

/* For only one bit-at-a-time "bitfields" the following
 * macros are easier, but note that, due to Endianness issues
 * there is no easy relationship between bit numbering with
 * these and with preceding functions.
 *
 * For gcc-Intel, correspondence can be obtained by setting
 *  BAJ to 7 (we don't bother because we don't rely on this
 *  correspondence).
 * If you chose to do so, remember that bitfield must then be
 *  whole number of unsigned longs
 */
#define	BAJ	0

#define	Bit_set(b, x)	(   ((char *)(b))[(x)>>3 ^ BAJ] |=   128u>>((x) & 7))
#define	Bit_clr(b, x)	(   ((char *)(b))[(x)>>3 ^ BAJ] &= ~(128u>>((x) & 7)))
#define	Bit_tst(b, x)	(!!(((char *)(b))[(x)>>3 ^ BAJ] &    128u>>((x) & 7)))


/* entry points in combinate.c */
long js_choose(int n, int m);
int js_kthcomb(long k, int n, int m, int result[]);
int *revdoor_init(int n, int m);
int revdoor_next(int arr[], int *in, int *out);

/* Entries in isprime.c */
int isprime(Uint32 x);
Uint32 nxtprime(Uint32 x);
