
/*
 * rummyexact.c  --  Calculate probability of Rummy Yarborough
 * Copyright (c) 2020 by James Dow Allen
 *
 * Results:

SIZHAND = 9
	6820068: 2 2 3 467 1217
	/
	3679075400: 2 2 2 5 5 7 11 13 17 23 47


	4262880: 2 2 2 2 2 3 5 83 107
	/
	15820024220: 2 2 5 7 11 13 17 23 43 47

.00185374510128278425
.00026946102867597253


	3*467*1217 / (2*5*5*7*11*13*17*23*47)


	2*2*2*3*83*107 / (7*11*13*17*23*43*47)
 */

#include	<stdlib.h>
#include	<stdio.h>

struct node {
	short s[4];
} Root;

#define	C_PER_S	13
#define	SIZHAND	9

long cnthand(struct node *p, int numc, int low)
{
	struct node This;
	int	j, suit, m, jj;
	long	tot = 0;

	if (numc == SIZHAND)
		return 1;
	for (j = low; j + SIZHAND - numc <= C_PER_S; j++)
	for (suit = 0; suit < 4; suit++) {
		m = p->s[suit];
		jj = 1 << j;
		if (m & jj >> 1) continue;
		if (m & jj >> 2) continue;
		This = *p;
		This.s[suit] |= jj;
		tot += cnthand(&This, numc + 1, j + 1);
	}
	return tot;
}

int main(int argc, char **argv) 
{
	int	i, j;
	long	tot = 0;

	for (j = 0; j < C_PER_S + 1 - SIZHAND; j++) {
		Root.s[0] = 1 << j;
		tot += cnthand(&Root, 1, j + 1);
	}

	printf("Tot = %d * 4 / C(52,%d)\n",
			tot, SIZHAND);
	exit(0);
}
