
/*
 * rummysim.c  --  Estimate probability of Rummy Yarborough
 * Copyright (c) 2020 by James Dow Allen
 */

#include	<stdlib.h>
#include	<stdio.h>
#include	<time.h>
#include	"j_random.h"

#define	SIZHAND	10

long long Deck[52];
#define	UPTR(y)	((short *)(Deck + (y)))

int 	simmer(void)
{
	short	suit[4], j;

	rand_llshuff(Deck, 52);
	suit[0] = suit[1] = suit[2] = suit[3] = 0;
	for (j = 0; j < SIZHAND; j++) {
		suit[0] |= UPTR(j)[0];
		suit[1] |= UPTR(j)[1];
		suit[2] |= UPTR(j)[2];
		suit[3] |= UPTR(j)[3];
	}
	if (suit[0] & suit[1]) return 0;
	if (suit[0] & suit[2]) return 0;
	if (suit[0] & suit[3]) return 0;
	if (suit[1] & suit[2]) return 0;
	if (suit[1] & suit[3]) return 0;
	if (suit[2] & suit[3]) return 0;
	for (j = 0; j < 4; j++) {
		if (suit[j] & suit[j] << 1) return 0;
		if (suit[j] & suit[j] << 2) return 0;
	}
	return 1;
}

void pcards(char sname, short sm)
{
	int	j;

	printf("%c", sname);
	for (j = 0; j < 13; j++) {
		if (sm & 1 << j) {
			printf("%c", "23456789TJQKA"[j]);
			sm ^= 1 << j;
		}
	}
}

void pcard(short *smp)
{
	if (smp[0])
		pcards('S', smp[0]);
	if (smp[1])
		pcards('H', smp[1]);
	if (smp[2])
		pcards('C', smp[2]);
	if (smp[3])
		pcards('D', smp[3]);
	printf(" ");
}

int main(int argc, char **argv) 
{
	int	i, j;

	rand_seed(time(0));
	for (j = 0; j < 52; j++) {
		UPTR(j)[j / 13] = 1 << j % 13;
	}

	for (i = 0; i < 30000000; i++) {
		if (simmer()) {
			for (j = 0; j < SIZHAND; j++)
				pcard(UPTR(j));
			printf("\n");
		}
	}
}
