
#include	"js_simple.h"

void *mustmalloc(size_t siz, char *msg)
{
	void *p;
	static unsigned long bytes_alloc = 0;

	if (siz == 0) {
		fprintf(stderr,
			"Mustmalloc has allocated %lu bytes total\n",
			bytes_alloc);
		return 0;
	}
	p = malloc(siz);
	if (!p) {
		fprintf(stderr, "Malloc(%lu) failed while allocating %s\n",
				(unsigned long)siz, msg ? msg : "?");
		mustmalloc(0, 0);
		exit(EXIT_MEMORY);
	}
	bytes_alloc += siz;
	return p;
}

/* G.C.D. functions */
GCD_Routine(js_gcd, unsigned long)
GCD_Routine(js_gcdbig, Bigint)

/* Table for fast bit counting */
#define	BC1(x)		x, x+1, x+1, x+2
#define	BC2(x)		BC1(x), BC1(x+1), BC1(x+1), BC1(x+2)
#define	BC3(x)		BC2(x), BC2(x+1), BC2(x+1), BC2(x+2)
#define	BC4(x)		BC3(x), BC3(x+1), BC3(x+1), BC3(x+2)
#define	BC5(x)		BC4(x), BC4(x+1), BC4(x+1), BC4(x+2)
static char BitCnt[1 << 11] = {
	BC5(0), BC5(1),
};

/* number of one-bits in g */
int js_bitcnt(unsigned long g)
{
	return	0
		+ BitCnt[g & 0x7ff]
		+ BitCnt[g>>11 & 0x7ff]
		+ BitCnt[g>>22 & 0x3ff];
}

int js_bbitcnt(Bigint g)
{
	return	js_bitcnt(g) +
		(BITS_IN_BIGINT > BITS_IN_ULONG
			? js_bitcnt(g >> BITS_IN_ULONG)
			: 0);
}

int locbit(int n, unsigned long m)
{
	int	i;

	for (i = 0; m; i++) {
		if (m & 1 && n-- == 0)
			return i;
		m >>= 1;
	}
	return -1;	/* error out */
}

/* index of n'th one-bit in m, counting from left */
int loclbit(int n, unsigned long m)
{
	int	i;
	unsigned long msk
		= (unsigned long)1 << BITS_IN_ULONG - 1;

	for (i = BITS_IN_ULONG - 1; i >= 0; i--) {
		if (msk & m && n-- == 0)
			return i;
		msk >>= 1;
	}
	return -1;	/* error out */
}

#define	SET_LBITS(p, k, v, t, z)			\
	(*(p) = *(p) & ((((t)1) << (z) -(k)) - 1)	\
		| (v) << (z) - (k))

#define	SET_GBITS(p, k, n, v, t, z)			\
	(*(p) = *(p) & (~((((t)1) << (z) - (n))	\
		  - (((t)1) << (z) - (n) - (k))))	\
		| (v) << (z) - (n) - (k))

/* set |bcnt| bits to |datum|, offset |boff| bits from *|p| */
#define	JSB_SET(name, typ, bper)				\
void name(typ *p, int bcnt, unsigned long boff, typ datum)	\
{								\
	int rbit;						\
								\
	p += boff / bper;					\
	boff %= bper;						\
	rbit = bcnt + boff;					\
	if (rbit > bper) {					\
		SET_GBITS(p, bper - boff, boff,			\
			datum >> rbit - bper,			\
			typ, bper);				\
		SET_LBITS(p + 1, rbit - bper,			\
			datum, typ, bper);			\
	} else if (boff) {					\
		SET_GBITS(p, bcnt, boff, datum, typ, bper);	\
	} else {						\
		SET_LBITS(p, bcnt, datum, typ, bper);		\
	}							\
}

/* return the |bcnt| bits at offset |boff| bits from *|p| */
#define	JSB_GET(name, typ, bper)				\
typ name(typ *p, int bcnt, unsigned long boff)			\
{								\
	int rbit;						\
	typ x;							\
								\
	p += boff / bper;					\
	rbit = bcnt + boff % bper;				\
	if (rbit > bper)					\
		x = p[0] << rbit - bper				\
			| p[1] >> 2*bper - rbit;		\
	else							\
		x = p[0] >> bper - rbit;			\
	if (bcnt < bper)					\
		x &= (((typ)1) << bcnt) - 1;			\
	return x;						\
}

JSB_SET(js_setbits, unsigned long, BITS_IN_ULONG)
JSB_GET(js_getbits, unsigned long, BITS_IN_ULONG)

/* Same functions for Bigint */
JSB_SET(js_bsetbits, Bigint, BITS_IN_BIGINT)
JSB_GET(js_bgetbits, Bigint, BITS_IN_BIGINT)

