/*
 * Convert a UTF-16 file or Ascii-8 to a UTF-8 file
 *	-E European-style Ascii-8
 *	-T Thai-style Ascii-8
 *	-R Random data
 *	[arg omitted] UTF-16
 */
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	"utf_class.h"

int	Verbose;

#define	IB_max	500000
int	IB_this, IB_that, IB_eof, IB_more;
unsigned short	IBuff[IB_max];
int	Lil_swapp;
int	IMode;

#define	OB_max	500000
int	OB_cnt, OB_more;
unsigned char	OBuff[500010];

void IB_prime(void)
{
	int	cnt;
	unsigned char *p, pv;
	int	j;
	static int beenhere = 0;

	IBuff[0] = IBuff[IB_this];
	IB_more += IB_this;
	IB_that -= IB_this;
	IB_this = 0;

	if (IB_eof == 0) {
		p = (unsigned char *)(IBuff + IB_that);
		cnt = beenhere++ ? IB_max - IB_that : 1;
		cnt = fread(p,
			sizeof *IBuff, 
			cnt,
			stdin);
		if (cnt == 0)
			IB_eof = 1;
		if (Lil_swapp) {
			for (j = 0; j < cnt; j++, p += 2) {
				pv = p[0];
				p[0] = p[1];
				p[1] = pv;
			}
		}
		IB_that += cnt;
	}
}

long Get_char(void)
{
	long	c, c2;
	static unsigned long rr = 0x7631,
		rrcounter = 1000000;

	switch (IMode) {
	case 't': case 'T':	/* Thai */
		c = getc(stdin);
		IB_more++;
		if (c == EOF)
			return EOF;
		else if (c < 0x80)
			return c;
		else
			return c + 0x0e00 - 0xa0;
	case 'e': case 'E':	/* European */
		c = getc(stdin);
		IB_more++;
		if (c == EOF)
			return EOF;
		else
			return c;
	case 'r': case 'R':	/* Random */
		if (rrcounter-- == 0)
			return EOF;
		do {
			rr *= 17;
			rr += 691;
			rr %= 0x00110000;
		} while (rr >> 11 == 0x36 >> 1);
		return rr;
	default:
		fprintf(stderr, "Unrecognized mode '%c'\n", IMode);
		exit(1);
	case 0:
		if (IB_that - IB_this <= 1)
			IB_prime();
		if (IB_that == IB_this)
			return EOF;
		c = IBuff[IB_this++];
		if (c >> 10 != 0x36)
			return c;
		if (IB_that == IB_this) {
			fprintf(stderr,
		"UTF-16 file ends with bad data *%lx*\n",
				c);
			return EOF;
		}
		c2 = IBuff[IB_this++];
		if (c2 >> 10 != 0x37) {
			fprintf(stderr,
		"Skipping bad *%lx %lx* in UTF-16 file IB=%d OB=%d\n",
				c, c2, IB_this + IB_more, OB_cnt + OB_more);
			return -2;
		}
		return 0x10000 + ((c & 0x3ff) << 10 | (c2 & 0x3ff));
	}
}

void Outp8(unsigned long val)
{

	if (val >= 0x00110000) {
		fprintf(stderr, "Suppressed bad UTF-16 *%lx* IB=%d OB=%d\n",
				val, IB_this + IB_more, OB_cnt + OB_more);
	} else if (val >= 0x00010000) {
		OBuff[OB_cnt++] = 0xf0 | val >> 18;
		OBuff[OB_cnt++] = 0x80 | val >> 12 & 0x3f;
		OBuff[OB_cnt++] = 0x80 | val >>  6 & 0x3f;
		OBuff[OB_cnt++] = 0x80 | val >>  0 & 0x3f;
	} else if (val >> 11 == 0x1b) {
		fprintf(stderr, "Suppressed bad UTF-16 *%lx* IB=%d OB=%d\n",
				val, IB_this + IB_more, OB_cnt + OB_more);
	} else if (val >= 0x00000800) {
		OBuff[OB_cnt++] = 0xe0 | val >> 12;
		OBuff[OB_cnt++] = 0x80 | val >>  6 & 0x3f;
		OBuff[OB_cnt++] = 0x80 | val >>  0 & 0x3f;
	} else if (val >= 0x00000080) {
		OBuff[OB_cnt++] = 0xc0 | val >>  6;
		OBuff[OB_cnt++] = 0x80 | val >>  0 & 0x3f;
	} else {
		OBuff[OB_cnt++] = val;
	}
	if (OB_cnt >= OB_max) {
		fwrite(OBuff, sizeof *OBuff, OB_cnt, stdout);
		OB_more += OB_cnt;
		OB_cnt = 0;
	}
}

int	BOM_mode = 2;	/* 2 always, 0 never, 1 according to input */

int main(int argc, char **argv)
{
	long	val;

	if (argc > 1 && !strcmp(argv[1], "-Bom")) {
		BOM_mode = 2;
		argc--;
		argv++;
	}
	if (argc > 1 && !strcmp(argv[1], "-NoBom")) {
		BOM_mode = 0;
		argc--;
		argv++;
	}
	if (argc > 1 && (argv[1][1] == 'v' || argv[1][1] == 'V')) {
		Verbose = 7;
		argc--;
		argv++;
	}
	if (argc > 1)
		IMode = argv[1][0];
	if (IMode == '-')
		IMode = argv[1][1];
	if (IMode && BOM_mode == 1)
		BOM_mode = 2;

	val = Get_char();
	if (IMode == 0) {
		if (val == 0xfffe) {
			Lil_swapp = 1;
			val = 0xfeff;
		} else if (val != 0xfeff) {
			fprintf(stderr, "No BOM at beginning of UTF-16 file\n");
		}

		if (val == 0xfeff) {
			if (BOM_mode)
				Outp8(0xfeff);
			val = Get_char();
		} else if (BOM_mode == 2) {
			Outp8(0xfeff);
		}
	} else if (BOM_mode == 2) {
		Outp8(0xfeff);
	}

	for (; val != EOF; val = Get_char()) {
		if (val >= 0) {
			if (Verbose & 2)
				utf_class(val);
			Outp8(val);
		}
	}
	if (OB_cnt)
		fwrite(OBuff, sizeof *OBuff, OB_cnt, stdout);
	if (Verbose & 2)
		utf_showhist();
	exit(0);
}
