
/* Convert a UTF-8 file to a UTF-16 file */
#include	<stdio.h>
#include	<stdlib.h>
#include	"utf_class.h"

/* To keep things slightly simpler, support only 20 MB files or smaller */
#define	IB_SIZE	(20 * 1024 * 1024)
unsigned char	IBuff[IB_SIZE];
int	IB_this, IB_that;

int	N_0, N_1, N_2, N_3, N_4;

long Get_utf8(void)
{
	int	c[4];

	c[0] = IBuff[IB_this++];
	if (c[0] < 0x80) {
		N_0 += 1;
		return c[0];
	} else if (c[0] < 0xc0) {
		N_1 += 1;
		return -2;
	} else if (c[0] < 0xe0) {
		N_2 += 1;
		if (IB_this + 1 > IB_that)
			return EOF;

#define	GET_T1(xx)					\
			xx = IBuff[IB_this++];		\
			if ((xx & 0xc0) != 0x80)	\
				return -3

		GET_T1(c[1]);
		return 0
			| (c[0] & 0x1f) << 6
			| (c[1] & 0x3f);
	} else if (c[0] < 0xf0) {
		N_3 += 1;
		if (IB_this + 2 > IB_that)
			return EOF;
		GET_T1(c[1]);
		GET_T1(c[2]);
		return 0
			| (c[0] & 0x0f) << 12
			| (c[1] & 0x3f) <<  6
			| (c[2] & 0x3f) <<  0;
	} else {
		N_4 += 1;
		if (IB_this + 3 > IB_that)
			return EOF;
		GET_T1(c[1]);
		GET_T1(c[2]);
		GET_T1(c[3]);
		return (0
			| (c[0] & 0x0f) << 18
			| (c[1] & 0x3f) << 12
			| (c[2] & 0x3f) <<  6
			| (c[3] & 0x3f) <<  0);
	}
}

void utferr(long val, int len, int beg)
{
	fprintf(stderr,
		"UTF-8 file in violation at offset %d, decoded value = *%x*\n",
		beg, val);
	if (len > 0) {
		fprintf(stderr,
			"\t(Length %d over-long)\n", len);
	} else if (len == 0) {
		fprintf(stderr,
			"\t(Illegal value)\n");
	} else if (len == -2) {
		fprintf(stderr,
			"\t(Unexpected 10xxxxxx byte)\n");
	} else if (len == -3) {
		fprintf(stderr,
			"\t(Missing 10xxxxxx byte)\n");
	}
}

int main(int argc, char **argv)
{
	int	len;
	long	val;
	unsigned short	s;

	IB_that = fread(IBuff, sizeof *IBuff, IB_SIZE, stdin);
	while (IB_this < IB_that) {
		len = - IB_this;
		val = Get_utf8();
		if (val == EOF) {
			fprintf(stderr, "Improper EOF in UTF-8 file.\n");
			break;
		}
		len += IB_this;
		if (val < 0) {
			utferr(0, val, IB_this - len);
			continue;
		}
		utf_class(val);
		s = val;
		if (val < 0x80) {
			if (len != 1)
				utferr(val, len, IB_this - len);
			fwrite((unsigned char *)&s, sizeof s, 1, stdout);
		} else if (val < 0x800) {
			if (len != 2)
				utferr(val, len, IB_this - len);
			fwrite((unsigned char *)&s, sizeof s, 1, stdout);
		} else if (val > 0x10ffff) {
			if (1)
				utferr(val, 0, IB_this - len);
		} else if (val > 0xffff) {
			if (len != 4)
				utferr(val, len, IB_this - len);
			val -= 0x10000;
			s = val >> 10;
			s |= 0xd800;
			fwrite((unsigned char *)&s, sizeof s, 1, stdout);
			s = val & 0x3ff;
			s |= 0xdc00;
			fwrite((unsigned char *)&s, sizeof s, 1, stdout);
		} else if (0xd800 <= val && val <= 0xdcff) {
			if (1)
				utferr(val, 0, IB_this - len);
		} else {
			if (len != 3)
				utferr(val, len, IB_this - len);
			fwrite((unsigned char *)&s, sizeof s, 1, stdout);
		}
	}
	fprintf(stderr,
		"Bytes Typ0 %d;  Typ1 %d;  Typ2 %d;  Typ3 %d;  Typ4 %d\n",
		N_0, N_1, N_2, N_3, N_4);
	utf_showhist();
	exit(0);
}
