
/* Copyright 2018 by James Dow Allen */

double gsolver(int numx, int numy, double *pay, float *xstrat, float *ystrat);

/*
 * Solve a game with numx columns and numy rows; return the value
 *	of the game to row-picker..
 * Ystrat will contain his optimal mix; Xstrat had column-picker's optimal mix.
 */
#include	<stdlib.h>
#include	<stdio.h>

#define	MAX_S	63

static double	A[MAX_S+1][MAX_S+1];
static int	Arowix[MAX_S];
static int	Acolix[MAX_S];
#define	INVBIT	0x4000
#define	EPSILON	.0000000002

static double InitA(double *payoff, int Wid, int Hei)
{
	int	i, j, k;
	double	vsave;

	for (i = 0; i < Hei; i++) {
		Arowix[i] = i;
		A[i][Wid] = 1;
	}
	for (j = 0; j < Wid; j++) {
		Acolix[j] = j | INVBIT;
		A[Hei][j] = -1;
	}
	A[Hei][Wid] = 0;
	for (i = 0; i < Hei; i++)
	for (j = 0; j < Wid; j++) {
		A[i][j] = *payoff++;
	}

	vsave = A[0][0];
	for (i = 0; i < Hei; i++)
	for (j = 0; j < Wid; j++)
		if (vsave > A[i][j])
			vsave = A[i][j];
	vsave -= 1;
	for (i = 0; i < Hei; i++)
	for (j = 0; j < Wid; j++)
		A[i][j] -= vsave;
	return vsave;
}

static void Solve(int Wid, int Hei)
{
	int	tt, i, j, p = -1, q;
	double	r, t1, t2, d, bnd;

	for (q = 0; q < Wid; ) {
		/* Loop to find pivot row */
		for (r = 0, i = 0; i < Hei; i++) {
			t1 = A[i][q];
			if (t1 <= EPSILON)
				continue;
			t2 = A[i][Wid];
			if (t2 <= 0) {
				p = i;
				break;
			} else if (t1 > t2*r) {
				p = i;
				r = t1 / t2;
			}
		}
		d = A[p][q];	/* Pivot */
		for (j = 0; j < Wid+1; j++)
			if (j != q)
				A[p][j] /= d;

		for (i = 0; i < Hei+1; i++)
		if (i != p) {
			for (j = 0; j < Wid+1; j++)
			if (j != q)
				A[i][j] -= A[i][q] * A[p][j];
			A[i][q] = -A[i][q] / d;
		}
		A[p][q] = 1 / d;
		tt = Arowix[p];
		Arowix[p] = Acolix[q];
		Acolix[q] = tt;
		q = Wid;
		bnd = - EPSILON;
		for (j = 0; j < Wid; j++)
		if (A[Hei][j] < bnd) {
			q = j;
			bnd = A[Hei][j];
		}
	}
}

static double ValOptStrat(float *xstrat, float *ystrat, int Wid, int Hei)
{
	int	i, j, ix;
	double	val;

	val = 1 / A[Hei][Wid];

	for (j = 0; j < Wid; j++) {
		if (Acolix[j] & INVBIT) {
			ix = Acolix[j] & ~INVBIT;
			xstrat[ix] = 0;
		} else {
			ix = Acolix[j];
			ystrat[ix] = A[Hei][j] * val;
		}
	}
	for (i = 0; i < Hei; i++) {
		if (Arowix[i] & INVBIT) {
			ix = Arowix[i] & ~INVBIT;
			xstrat[ix] = A[i][Wid] * val;
		} else {
			ix = Arowix[i];
			ystrat[ix] = 0;
		}
	}
	return val;
}

static void chk(int numx, int numy, double *pay,
		float *xstrat, float *ystrat, double gameval);

double gsolver(int numx, int numy, double *pay, float *xstrat, float *ystrat)
{
	int	i;
	double	gameval;

	gameval = InitA(pay, numx, numy);
	Solve(numx, numy);
	gameval += ValOptStrat(xstrat, ystrat, numx, numy);
	chk(numx, numy, pay, xstrat, ystrat, gameval);
	return gameval;
}

static double dotprod(double *a, float *b, int stride, int size)
{
	double	res = 0;

	while (size--) {
		res += *a * *b++;
		a += stride;
	}
	return res;
}

static void chk(int numx, int numy, double *pay,
			float *xstrat, float *ystrat, double gameval)
{
	int	x, y;
	double	tval;

	for (x = 0; x < numx; x++) {
		tval = dotprod(pay + x, ystrat, numx, numy);
		if (tval < gameval - 0.0000002)
			fprintf(stderr,
				"Try x: %f > %f\n", tval, gameval);
	}
	for (y = 0; y < numy; y++) {
		tval = dotprod(pay + y*numx, xstrat, 1, numx);
		if (tval > gameval + 0.0000002)
			fprintf(stderr,
				"Try y: %f > %f\n", tval, gameval);
	}
}
