
#include	<stdio.h>
#include	<stdlib.h>
#include	"rat_arith.h"

/*
 * Warning:  These functions do NOT check for overflow
 * Unless uou define FAST.
 */

#define EXIT_GOOD       0
#define EXIT_EXCEED     17
#define EXIT_USAGE      64      /* command line usage error */

#ifdef	FAST
#define	FAST

Add(x, y)	((x) + (y))
Mpy(x, y)	((x) * (y))

#else

static long long Add(long long x, long long y)
{
	int	signx = x < 0;
	int	signy = y < 0;
	long long sum = x + y;
	if (signx != signy || x == 0 || y == 0)
		return sum;
	if (sum < 0 && signx || sum > 0 && !signx)
		return sum;
	fprintf(stderr, "Overflow adding %lld and %lld\n", x, y);
	exit(EXIT_EXCEED);
}

static long long Mpy(long long x, long long y)
{
	int	signx;
	long long prod;

	if (x == 0 || y == 0)
		return 0;
	if (x < 0) {
		signx = 1;
		x = -x;
	} else {
		signx = 0;
	}
	if (y < 0) {
		signx ^= 1;
		y = -y;
	}
	prod = x * y;
	if (prod >= x && prod >= y)
		return signx ? -prod : prod;
	fprintf(stderr, "Overflow multiplying %lld and %lld\n", x, y);
	exit(EXIT_EXCEED);
}

#endif

Ratio set_rat(long long anum, long long aden)
{
	Ratio t;
	if (aden < 0) {
		t.num = -anum;
		t.den = -aden;
	} else {
		t.num = anum;
		t.den = aden;
	}
	return t;
}

char *pri_rat(Ratio a)
{
	static char buff[10][100];
	static int bix = 0;

	bix = (bix + 1) % 10;
	if (a.den < 0)
		sprintf(buff[bix], "%lld/%lld", -a.num, -a.den);
	else
		sprintf(buff[bix], "%lld/%lld", a.num, a.den);
	return buff[bix];
}

long long gcd(long long a, long long b)
{

	if (a < 0) a = -a;
	if (b < 0) b = -b;
	if (a == 0)
		return b ? b : 1;
	if (b == 0)
		return a;
	while (1)
		if (a <= b && (b %= a) < 2)
			return b ? 1 : a;
		else if ((a %= b) < 2)
			return a ? 1 : b;
}

Ratio in_add_rat(long long anum, long long aden,
			long long bnum, long long bden)
{
	long long t1, t2, gg;
	Ratio t;

	t1 = Add(Mpy(anum, bden), Mpy(bnum, aden));
	t2 = Mpy(aden, bden);
	gg = gcd(t1, t2);
	t.num = t1 / gg;
	t.den = t2 / gg;
	return t;
}

Ratio recip_rat(Ratio a)
{
	Ratio t;
	if (a.num < 0) {
		t.num = - a.den;
		t.den = - a.num;
	} else {
		t.num = a.den;
		t.den = a.num;
	}
	return t;
}

Ratio in_mul_rat(long long anum, long long aden,
			long long bnum, long long bden)
{
	long long t1, t2, gg;
	Ratio t;

	gg = gcd(anum, bden);
	anum /= gg;
	bden /= gg;
	gg = gcd(aden, bnum);
	aden /= gg;
	bnum /= gg;
	t.num = Mpy(anum, bnum);
	t.den = Mpy(aden, bden);
	if (t.num == 0) {
		t.den = 1;
	} else if (t.den < 0) {
		t.num *= -1;
		t.den *= -1;
	}
	return t;
}

Ratio add_rat(Ratio a, Ratio b)
{
	return in_add_rat(a.num, a.den, b.num, b.den);
}

Ratio sub_rat(Ratio a, Ratio b)
{
	return in_add_rat(a.num, a.den, - b.num, b.den);
}

Ratio mul_rat(Ratio a, Ratio b)
{
	return in_mul_rat(a.num, a.den, b.num, b.den);
}

Ratio mul_const(Ratio a, long long x)
{
	return in_mul_rat(a.num, a.den, x, 1);
}

Ratio div_rat(Ratio a, Ratio b)
{
	return in_mul_rat(a.num, a.den, b.den, b.num);
}

int gt_rat(Ratio a, Ratio b)
{
	Ratio t = sub_rat(a, b);
	return t.num * t.den > 0;
}

#ifdef	CFRACT
Ratio Rat1, Rat2, Rat3;

int main(int argc, char **argv)
{
	long long s;

	setlinebuf(stdout);
	if (argc != 3) {
		printf("Usage: cf #num #den\n");
		printf("   -- express a fraction in CF form\n");
		exit(EXIT_USAGE);
	}
	Rat1.num = atoll(argv[1]);
	Rat1.den = atoll(argv[2]);
	if (Rat1.num < 0) Rat1.num = - Rat1.num;
	if (Rat1.den < 0) Rat1.den = - Rat1.den;
	if (Rat1.den == 0) Rat1.den = 1;
	printf(" %lld / %lld = ", Rat1.num, Rat1.den);

	s = Rat1.num / Rat1.den;
	printf("%lld ", s);
	Rat1.num -= s * Rat1.den;
	while (Rat1.num && Rat1.den) {
		s = Rat1.den / Rat1.num;
		printf("+1 / %lld", s);
		if (Rat1.num == 1) {
			break;
		}
		Rat1 = recip_rat(Rat1);
		Rat1.num -= s * Rat1.den;
	}
	printf("\n");
	exit(EXIT_GOOD);
}
#endif
