

/*
 * Solve a game with numx columns and numy rows; return the value
 *	of the game to row-picker..
 * Ystrat will contain his optimal mix; Xstrat had column-picker's optimal mix.
 */
#include	<stdlib.h>
#include	<stdio.h>
#include	"rat_arith.h"

#define	MAX_S	63

static Ratio	A[MAX_S+1][MAX_S+1];
static int	Arowix[MAX_S];
static int	Acolix[MAX_S];
#define	INVBIT	0x4000

Ratio gsolver(int numx, int numy, Ratio *pay, Ratio *xstrat, Ratio *ystrat);
Ratio igsolver(int numx, int numy, int *pay, Ratio *xstrat, Ratio *ystrat);

static Ratio InitA(Ratio *payoff, int Wid, int Hei)
{
	int	i, j, k;
	Ratio	vsave;

	for (i = 0; i < Hei; i++) {
		Arowix[i] = i;
		A[i][Wid] = set_rat(1, 1);
	}
	for (j = 0; j < Wid; j++) {
		Acolix[j] = j | INVBIT;
		A[Hei][j] = set_rat(-1, 1);
	}
	A[Hei][Wid] = set_rat(0, 1);
	for (i = 0; i < Hei; i++)
	for (j = 0; j < Wid; j++) {
		A[i][j] = *payoff++;
	}

	vsave = A[0][0];
	for (i = 0; i < Hei; i++)
	for (j = 0; j < Wid; j++)
		if (gt_rat(vsave, A[i][j]))
			vsave = A[i][j];
	vsave = sub_rat(vsave, set_rat(1, 1));
	for (i = 0; i < Hei; i++)
	for (j = 0; j < Wid; j++)
		A[i][j] = sub_rat(A[i][j], vsave);
	return vsave;
}

static void Solve(int Wid, int Hei)
{
	int	tt, i, j, p = -1, q;
	Ratio	r, t1, t2, d, bnd;

	for (q = 0; q < Wid; ) {
		/* Loop to find pivot row */
		r = set_rat(0, 1);
		for (i = 0; i < Hei; i++) {
			t1 = A[i][q];
			if (t1.num * t1.den <= 0)
				continue;
			t2 = A[i][Wid];
			if (t2.num * t2.den <= 0) {
				p = i;
				break;
			} else if (gt_rat(t1, mul_rat(t2, r))) {
				p = i;
				r = div_rat(t1, t2);
			}
		}
		d = A[p][q];	/* Pivot */
		for (j = 0; j < Wid+1; j++) {
			if (j != q) {
				A[p][j] = div_rat(A[p][j], d);
			}
		}

		for (i = 0; i < Hei+1; i++)
		if (i != p) {
			for (j = 0; j < Wid+1; j++)
			if (j != q)
				A[i][j] = sub_rat(A[i][j],
					mul_rat(A[i][q], A[p][j]));
			d.num *= -1;
			A[i][q] = div_rat(A[i][q], d);
			d.num *= -1;
		}
		A[p][q].num = d.den;
		A[p][q].den = d.num;
		tt = Arowix[p];
		Arowix[p] = Acolix[q];
		Acolix[q] = tt;
		q = Wid;
		bnd = set_rat(0, 1);
		for (j = 0; j < Wid; j++)
		if (gt_rat(bnd, A[Hei][j])) {
			q = j;
			bnd = A[Hei][j];
		}
	}
}

static Ratio ValOptStrat(Ratio *xstrat, Ratio *ystrat, int Wid, int Hei)
{
	int	i, j, ix;
	Ratio	val;

	val.num = A[Hei][Wid].den;
	val.den = A[Hei][Wid].num;

	for (j = 0; j < Wid; j++) {
		if (Acolix[j] & INVBIT) {
			ix = Acolix[j] & ~INVBIT;
			xstrat[ix] = set_rat(0, 1);
		} else {
			ix = Acolix[j];
			ystrat[ix] = mul_rat(A[Hei][j], val);
		}
	}
	for (i = 0; i < Hei; i++) {
		if (Arowix[i] & INVBIT) {
			ix = Arowix[i] & ~INVBIT;
			xstrat[ix] = mul_rat(A[i][Wid], val);
		} else {
			ix = Arowix[i];
			ystrat[ix] = set_rat(0, 1);
		}
	}
	return val;
}

static void chk(int numx, int numy, Ratio *pay,
		Ratio *xstrat, Ratio *ystrat, Ratio gameval);

Ratio gsolver(int numx, int numy, Ratio *pay, Ratio *xstrat, Ratio *ystrat)
{
	int	i;
	Ratio	gameval;

	gameval = InitA(pay, numx, numy);

	Solve(numx, numy);

	gameval = add_rat(gameval, ValOptStrat(xstrat, ystrat, numx, numy));
	chk(numx, numy, pay, xstrat, ystrat, gameval);
	return gameval;
}

Ratio igsolver(int numx, int numy, int *pay, Ratio *xstrat, Ratio *ystrat)
{
	Ratio Rpay[144];
	int	i;

	for (i = 0; i < numx * numy; i++) {
		Rpay[i].num = pay[i];
		Rpay[i].den = 1;
	}
	return gsolver(numx, numy, Rpay, xstrat, ystrat);
}

static Ratio dotprod(Ratio *a, Ratio *b, int stride, int size)
{
	Ratio	res = set_rat(0, 1);

	while (size--) {
		res = add_rat(res, mul_rat(*a, *b));
		a += stride;
		b += 1;
	}
	return res;
}

static void chk(int numx, int numy, Ratio *pay,
			Ratio *xstrat, Ratio *ystrat, Ratio gameval)
{
	int	x, y;
	Ratio	tval;

	for (x = 0; x < numx; x++) {
		tval = dotprod(pay + x, ystrat, numx, numy);
	}
	for (y = 0; y < numy; y++) {
		tval = dotprod(pay + y*numx, xstrat, 1, numx);
	}
}
