
/* Copright 2018 by James Dow Allen */

#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	"j_random.h"
#include	"rat_arith.h"

#define EXIT_GOOD       0
#define EXIT_EXCEED     17      /* method limit reached */

Ratio igsolver(int numx, int numy, int *pay, Ratio *xstrat, Ratio *ystrat);
unsigned long long canon_sig(int siz, int *bmat);

#ifndef	JSIZ
#define	JSIZ	7
#endif

#define	TRISIZ	(JSIZ * (JSIZ-1) / 2)

#define	NUMS_X	JSIZ
#define	NUMS_Y	JSIZ

int	pays[NUMS_Y * NUMS_X] = {
#if	JSIZ == 7
	 0, -1, -1, -1, -1, -1, -1,
	 1,  0, -1, -1, -1, -1, -1,
	 1,  1,  0, -1, -1, -1, -1,
	 1,  1,  1,  0, -1,  1, -1,
	 1,  1,  1,  1,  0, -1, -1,
	 1,  1,  1, -1,  1,  0, -1,
	 1,  1,  1,  1,  1,  1,  0,
#endif
#if	JSIZ == 9
	 0,  1,  1, -1, -1, -1,  1,  1, -1,
	-1,  0,  1,  1,  1, -1,  1, -1,  1,
	-1, -1,  0,  1, -1,  1,  1,  1, -1,
	 1, -1, -1,  0, -1,  1,  1, -1,  1,
	 1, -1,  1,  1,  0, -1, -1, -1, -1,
	 1,  1, -1, -1,  1,  0,  1, -1, -1,
	-1, -1, -1, -1,  1, -1,  0,  1,  1,
	-1,  1, -1,  1,  1,  1, -1,  0,  1,
	 1, -1,  1, -1,  1,  1, -1, -1,  0,
#endif
};

/* Next will be set same as pays, but as single bits
 *   with both 0 and -1 mapped to 0.
 * It will then be used as work area by canon_sig().
 */
int	Bmatr[JSIZ];

void setmat(long long msk)
{
	int	i, j, mshft, t;

	mshft = TRISIZ;
	for (i = 0; i < NUMS_Y; i++) {
		Bmatr[i] = 0;
		for (j = 0; j < NUMS_X; j++) {
			t = pays[i * NUMS_X + j] =
				(i == j) ? 0
				: (i > j) ? -pays[j * NUMS_X + i]
				: (1 & msk >> --mshft) ? 1 : -1;
			if (t > 0)
				Bmatr[i] |= 1 << NUMS_X-1-j;
		}
	}
}

#define	d2i(x)	"-0+"[1 + x]

struct Nprint {
	Ratio	prob;
	int	nwin;
} Np[NUMS_Y];

int npcmp(const void *p1, const void *p2)
{
	Ratio ip1 = ((struct Nprint *)p1)->prob;
	Ratio ip2 = ((struct Nprint *)p2)->prob;
	int iw1 = ((struct Nprint *)p1)->nwin;
	int iw2 = ((struct Nprint *)p2)->nwin;

	if (gt_rat(ip2, ip1)) return 1;
	if (gt_rat(ip1, ip2)) return -1;
	if (iw1 < iw2) return 1;
	if (iw1 > iw2) return -1;
	return 0;
}

int newsetsyn(Ratio *ymx)
{
	int	i, j, k;
	Ratio	sump = set_rat(0, 1);
	long long the_lcm = 1, t;

	for (i = 0; i < NUMS_Y; i++) {
		for (j = k = 0; j < NUMS_X; j++) {
			k += pays[i * NUMS_Y + j] > 0;
		}
		Np[i].nwin = k;
		Np[i].prob = ymx[i];
		sump = add_rat(sump, Np[i].prob);
		the_lcm *= ymx[i].den / gcd(the_lcm, ymx[i].den);
	}
	if (sump.den != sump.num)
		printf("Gak! Probabilities don't sum.\n");
	for (i = 0; i < NUMS_Y; i++) {
		t = the_lcm / Np[i].prob.den;
		Np[i].prob.num *= t;
		Np[i].prob.den = the_lcm;
	}
	qsort(Np, NUMS_Y, sizeof *Np, npcmp);
	printf("(");
	for (i = 0; i < NUMS_Y; i++) {
		printf("%lld_%d:, ", Np[i].prob.num, Np[i].nwin);
	}
	printf(") &#247; %lld\n", the_lcm);
}

/* A hash table to record previous discoveries */

#if	0
#define	NUMEL	990001
#define	NUMHCAB	1
#else
/* The above (NUMHCAB 1) works finem but let's put code through paces. */
#define	NUMEL	34141
#define	NUMHCAB	29
#endif

struct hashe {
	long long valu;
} Hashe[NUMHCAB][NUMEL];

#define	HT_EMPTY(np)		((np)->valu == 0)
#define HT_MATCH(np, kkk)	((np)->valu == (kkk))
#define HT_SETIT(np, key)	((np)->valu = 1 + (key) / NUMHCAB)

/* The above customizes a particular hash table,
 *  Allowing us to use a small generic source function:
 */
#include	"ezhash.c"


void solvprint(int *gpay)
{
	Ratio	xmix[NUMS_X], ymix[NUMS_Y];
	Ratio	gameval;
	int	i, j;
	long long sigsig;
	struct hashe *np;

	gameval = igsolver(NUMS_X, NUMS_Y, gpay, xmix, ymix);

#if	0
	/* game value always zero for RPS or other symmetric games */
	printf("Game value = %s\n", pri_rat(gameval));
#endif
	if (gameval.num != 0)
		printf("Gak! Game value isn't zero.\n");

	/* Reject degenerate cases */
	for (i = 0; i < NUMS_Y; i++)
		if (ymix[i].num == 0)
			return;

	sigsig = canon_sig(NUMS_X, Bmatr);

	np = findnode(sigsig);
	if (HT_EMPTY(np)) {
		/* New node */
		HT_SETIT(np, sigsig);
	} else {
		/* Re-encountered old node */
		if (1)
			return;
		if (0)
		for (i = 0; i < NUMS_Y; i++) {
			for (j = 0; j < NUMS_X; j++)
				printf("%c", d2i(gpay[i * NUMS_X + j]));
			printf(i == NUMS_Y - 1 ? " = " : " / ");
		}
		printf(" Sig=%013llx re-encountered\n", sigsig);
		return;
	}

	for (i = 0; i < NUMS_Y; i++) {
		for (j = 0; j < NUMS_X; j++)
			printf("%c", d2i(gpay[i * NUMS_X + j]));
		printf(i == NUMS_Y - 1 ? " = " : " / ");
	}

	printf(" Sig=%013llx ", sigsig);
	newsetsyn(ymix);

	/* Unzero next for fuller report */
	if (0) {
		printf("Colgirl's strategy is ");
		for (i = 0; i < NUMS_X; i++)
			printf(" %s", pri_rat(xmix[i]));
		printf("\nRowboy's strategy is ");
		for (i = 0; i < NUMS_Y; i++)
			printf(" %s", pri_rat(ymix[i]));
		printf("  Value = %s\n", pri_rat(gameval));
	}
}

int main(int argc, char **argv)
{
	long long	j, j2;
	int 	do_random = 0;

	setlinebuf(stdout);

	if (JSIZ == 11 || argv[1] && !strcmp(argv[1], "-R"))
		do_random = 1;
	if (JSIZ > 11 || JSIZ < 3 || !(JSIZ & 1)) {
		printf("Only Jsiz of 3, 5, 7, 9, 11 are permitted\n");
		exit(EXIT_EXCEED);
	}

	/* Unzero next to solve a particular game hand-entered into pays[] */ 
#if	0
	solvprint(pays);
	exit(EXIT_GOOD);
#endif

	if (do_random) {
		if (argv[1] && argv[2])
			rand_seed(atoi(argv[2]));
		do {
#if	(JSIZ >= 7)
			j = rand_index(1 << 21);
#else
			j = rand_index(1 << TRISIZ);
#endif
#if	(JSIZ >= 9)
			j2 = rand_index(1 << 15);
#else
			j2 = 0;
#endif
#if	(JSIZ >= 11)
			j2 = j2 << 19 | rand_index(1 << 19);
#endif
			setmat(0x000 | j | j2 << 21);
			solvprint(pays);
		} while (1);
	}

#if	JSIZ == 9
	/* Following gets all 792 when JSIZ == 9 */
	for (j = 0; j < 1 << 21; j += 0x400) {
		for (j2 = 6 << 12; j2 < 7 << 12; j2 += 1) {
			setmat(0x277 | j | j2 << 21);
			solvprint(pays);
			setmat(0x2f1 | j | j2 << 21);
			solvprint(pays);
		}
	}
#else
	for (j = 0; j < ((long long) 1) << TRISIZ; j += 1) {
		setmat(j);
		solvprint(pays);
	}
#endif
	exit(EXIT_GOOD);
}
