
#include	<stdlib.h>
#include	<stdio.h>
#include	<math.h>
#include	<string.h>

#define	MAIN_MODULE
#include	"js_simple.h"

void jacobi(double *a, int n, double *d, double *v, int *nrot);

int	verbose;

/* Here come the determinant-finding routines...  */
#define	DET2(a00, a01, a10, a11)				\
	  (a00 * a11 - a01 * a10)

#define	DET3(a00, a01, a02, a10, a11, a12, a20, a21, a22)	\
	 (a00 * a11 * a22 + a01 * a12 * a20 + a02 * a10 * a21	\
	- a02 * a11 * a20 - a01 * a10 * a22 - a00 * a12 * a21)

#define	MDET3(m, a, b, c)		\
		DET3(m[1][a], m[1][b], m[1][c],	\
		m[2][a], m[2][b], m[2][c],	\
		m[3][a], m[3][b], m[3][c])

#define	MDET4(m)		\
		(m[0][0] * MDET3(m, 1,2,3) -	\
		 m[0][1] * MDET3(m, 0,2,3) +	\
		 m[0][2] * MDET3(m, 0,1,3) -	\
		 m[0][3] * MDET3(m, 0,1,2))

/* Set b = a^-1 ;  both are 3x3 matrices */
void invert33(double (*a)[3], double (*b)[3])
{
	double	rdeter;

	rdeter = 1.0 /
	   DET3(a[0][0], a[1][0], a[2][0],
		a[0][1], a[1][1], a[2][1],
		a[0][2], a[1][2], a[2][2]);
#define	CTR(x,n)		((x) < n ? (x) : (x) - n)
#define	CT_SETI(x, y)					\
	b[y][x] = rdeter * DET2(		\
		a[CTR(x+1, 3)][CTR(y+1, 3)],	\
		a[CTR(x+2, 3)][CTR(y+1, 3)],	\
		a[CTR(x+1, 3)][CTR(y+2, 3)],	\
		a[CTR(x+2, 3)][CTR(y+2, 3)])

	CT_SETI(0, 0); CT_SETI(0, 1); CT_SETI(0, 2);
	CT_SETI(1, 0); CT_SETI(1, 1); CT_SETI(1, 2);
	CT_SETI(2, 0); CT_SETI(2, 1); CT_SETI(2, 2);
}

/* Set b = a^-1 ;  both are 4x4 matrices */
void invert44(double (*a)[4], double (*b)[4])
{
	double	rdeter;

	rdeter = 1.0 / MDET4(a);
#define	CT4_SETI(x, y)					\
	b[y][x] = (((x + y) & 1) ? -1 : 1) * rdeter * DET3(		\
		a[CTR(x+1, 4)][CTR(y+1, 4)],	\
		a[CTR(x+2, 4)][CTR(y+1, 4)],	\
		a[CTR(x+3, 4)][CTR(y+1, 4)],	\
		a[CTR(x+1, 4)][CTR(y+2, 4)],	\
		a[CTR(x+2, 4)][CTR(y+2, 4)],	\
		a[CTR(x+3, 4)][CTR(y+2, 4)],	\
		a[CTR(x+1, 4)][CTR(y+3, 4)],	\
		a[CTR(x+2, 4)][CTR(y+3, 4)],	\
		a[CTR(x+3, 4)][CTR(y+3, 4)])

	CT4_SETI(0, 0); CT4_SETI(0, 1); CT4_SETI(0, 2); CT4_SETI(0, 3);
	CT4_SETI(1, 0); CT4_SETI(1, 1); CT4_SETI(1, 2); CT4_SETI(1, 3);
	CT4_SETI(2, 0); CT4_SETI(2, 1); CT4_SETI(2, 2); CT4_SETI(2, 3);
	CT4_SETI(3, 0); CT4_SETI(3, 1); CT4_SETI(3, 2); CT4_SETI(3, 3);
}

/* b = transpose a */
void transpose(double *a, double *b, int AWID, int AHEI)
{
	int	i, j;

	for (i = 0; i < AWID; i++)
	for (j = 0; j < AHEI; j++)
		b[i*AHEI + j] = a[j*AWID + i];
}

void jtrotate(double *a, double *b, int size, double tang)
{
	int	i;
	double	ta, tb, m1, m2;

	m2 = 1 / sqrt(1 + tang * tang);
	m1 = m2 * tang;

	for (i = 0; i < size; ++i, ++a, ++b) {
		ta = *a;
		tb = *b;
		*a = ta * m2 + tb * m1;
		*b = tb * m2 - ta * m1;
	}
}

/*
 * Mix vectors a[] and b[] together; ie rotate by arctan(mix / (1 - mix))
 */
void jrotate(double *a, double *b, int size, double mix)
{
	double	tang;

	tang = mix > 0 ? mix / (1 - mix) : mix / (1 + mix);
	jtrotate(a, b, size, tang);
}

/* b = sub_matrix of a */
void submat(double *a, double *b, int AWID, int BWID, int BHEI)
{
	int	i, j;

	for (i = 0; i < BHEI; i++, a += AWID)
	for (j = 0; j < BWID; j++, b++)
		b[0] = a[j];
}

void normalize(double *a, int AWID, int AHEI)
{
	int	i, j;
	double	sumsq;

	for (i = 0; i < AHEI; i++) {
		for (sumsq = j = 0; j < AWID; j++)
			sumsq += a[i*AWID + j] * a[i*AWID + j];
		sumsq = sqrt(sumsq);
		for (j = 0; j < AWID; j++)
			a[i*AWID + j] /= sumsq;
	}
}

double	ddotprod(double *a, double *b, int cnt, int binc)
{
	double	sum;

	for (sum = 0; cnt--; a += 1, b += binc)
		sum += *a * *b;
	return sum;
}

/* c = a * b
	*a ::	wid = AWID hei = AHEI
	*b ::	wid = BWID hei = AWID
	*c ::	wid = BWID hei = AHEI
 */
void multiply(double *a, double *b, double *c, int AWID, int AHEI, int BWID)
{
	int	i, j;

	for (i = 0; i < AHEI; i++, a += AWID)
	for (j = 0; j < BWID; j++, c++)
		*c = ddotprod(a, b + j, AWID, BWID);
}

/* One idiom occurs so frequently, se have a special entry for it
 *   Set C to the MxM matrix A B A^T where B is NxN 
	*a ::	wid = BWID hei = AHEI
	*b ::	wid = BWID hei = BWID
	*c ::	wid = AHEI hei = AHEI
 */
void abat(double *a, double *b, double *c, int AHEI, int BWID)
{
	double	*at, *bat;

	at = (double *)mustmalloc(AHEI * BWID * sizeof(double), "Alloc AT");
	bat = (double *)mustmalloc(AHEI * BWID * sizeof(double), "Alloc BAT");
	transpose(a, at, BWID, AHEI);
	multiply(b, at, bat, BWID, BWID, AHEI);
	multiply(a, bat, c, BWID, AHEI, AHEI);
	free(at);
	free(bat);
}

void sodump(double *a, int AWID, int AHEI)
{
	int	i, j;

	for (i = 0; i < AHEI; i++) {
		for (j = 0; j < AWID; j++)
			printf(" %9.6f", a[i*AWID + j]);
		printf("\n");
	}
}

void aprint(char *msg, double *a, int AWID, int AHEI)
{
	int	i, j;

	fprintf(stderr, "%s : \n", msg);
	for (i = 0; i < AHEI; i++) {
		for (j = 0; j < AWID; j++)
			fprintf(stderr, " %9.6f", a[i*AWID + j]);
		fprintf(stderr, "\n");
	}
}

double dmedian3(double a, double b, double c)
{
	double	d;

	if (a < b)
		if (b < c) d = b;
		else if (a < c) d = c;
		else d = a;
	else
		if (b > c) d = b;
		else if (a > c) d = c;
		else d = a;
	return d;
}

double dmedian5(double a, double b, double c, double d, double e)
{
	double	t;

	if (a < b)
		t = a, a = b, b = t;
	if (c < d)
		t = c, c = d, d = t;
	t = dmedian3(e, a > c ? c : a, b > d ? b : d);
	return t;
}

/*
 * Now some general-purpose matrix-inversion code,
 * more-or-less stolen from Press et al, _Numerical Recipes in C_.
 *
 * Press et al seem to think this is the best way to invert it;
 * major inefficiency will be in my godforsaken "Matr()" macro.
 *
 * Instead of ABENDing if the matrix appears singular, we return `0'.
 * Hopefully the caller will check for this.   0.01 :-}
 */

#define	TINY	1.0e-20
#define	MAXDIM	100	/* largest size we sill attempt */

static int indx[MAXDIM];

#define	Matr(arr, row, col)	\
	((arr)[(col) + (row) * size])

/* Ludcmp returns -1, 0, or +1 ... 0 if singular */
static int ludcmp(double *in, int size)
{
	double	*inp, sum, tmp, biggy, vv[MAXDIM];
	int	imax = 0, i, j, k, icparity = 0;

	for (inp = in, i = 0; i < size; i++) {
		for (biggy = j = 0; j < size; j++, inp++) {
			tmp = fabs(*inp);
			if (tmp > biggy)
				biggy = tmp;
		}
		if (biggy < TINY) {
			fprintf(stderr, "Ludcmp: singular\n");
			return 0;
		}
		vv[i] = 1 / biggy;
	}
	for (j = 0; j < size; j++) {
		for (i = 0; i < j; i++) {
			sum = Matr(in, i, j);
			for (k = 0; k < i; k++)
				sum -= Matr(in, i, k) * Matr(in, k, j);
			Matr(in, i, j) = sum;
		}
		for (biggy = 0, i = j; i < size; i++) {
			sum = Matr(in, i, j);
			for (k = 0; k < j; k++)
				sum -= Matr(in, i, k) * Matr(in, k, j);
			Matr(in, i, j) = sum;
			tmp = vv[i] * fabs(sum);
			if (tmp >= biggy) {
				biggy = tmp;
				imax = i;
			}
		}
		if (j != imax) {
			for (k = 0; k < size; k++) {
				tmp = Matr(in, imax, k);
				Matr(in, imax, k) = Matr(in, j, k);
				Matr(in, j, k) = tmp;
			}
			icparity ^= 1;
			vv[imax] = vv[j];
		}
		indx[j] = imax;
		tmp = Matr(in, j, j);
		if (fabs(tmp) < TINY)
			Matr(in, j, j) = tmp = TINY;
		for (i = j+1; i < size; i++)
			Matr(in, i, j) /= tmp;
	}
#if	0
	for (j = 0; j < size; j++) {
		fprintf(stderr, "Index %d = %d\n", j, indx[j]);
	}
#endif
	return icparity ? -1 : 1;
}

/* Given a size-X-size matrix, destroy it and return its determinant */
double determ(double *in, int size)
{
	double	determ;
	int	i;

	determ = ludcmp(in, size);
	for (i = 0; i < size; i++)
		determ *= Matr(in, i, i);
	return determ;
}

/* Solve  A X = B; in is the LU decomp of A */
static void lubksb(double *in, double *b, int size)
{
	int	i, ip, j, ii = -1;
	double	sum;

	for (i = 0; i < size; i++) {
		ip = indx[i];
		sum = b[ip];
		b[ip] = b[i];
		if (ii >= 0)
			for (j = ii; j < i; j++)
				sum -= Matr(in, i, j) * b[j];
		else if (sum)
			ii = i;
		b[i] = sum;
	}
	for (i = size - 1; i >= 0; i--) {
		sum = b[i];
		for (j = i + 1; j < size; j++)
			sum -= Matr(in, i, j) * b[j];
		b[i] = sum / Matr(in, i, i);
	}
}

/* Same as matinvert but destroys `in' */
int do_invert(double *in, double *out, int size)
{
	double	acol[MAXDIM];
	int	i, j;

	if (ludcmp(in, size) == 0)
		return 0;
	for (j = 0; j < size; j++) {
		for (i = 0; i < size; i++)
			acol[i] = 0;
		acol[j] = 1;
		lubksb(in, acol, size);
		for (i = 0; i < size; i++)
			Matr(out, i, j) = acol[i];
	}
	return 1;
}

/* Finally!  ... invert a size-X-size matrix; return 1 if we did */
int matinvert(double *in, double *out, int size)
{
	double	*work;
	int	i, j;

	work = (double *)mustmalloc(size * size * sizeof(double), "Alloc WORK");
	for (i = 0; i < size; i++)
	for (j = 0; j < size; j++)
		Matr(work, i, j) = Matr(in, i, j);
	i = do_invert(work, out, size);
	free(work);
	return i;
}

/*
 * Now some eigensystem code, again
 * more-or-less stolen from Press et al, _Numerical Recipes in C_.
 */

static int	jix[MAXDIM];
static double	eigenval[MAXDIM];

int jixcomp(const void *j1, const void *j2)
{
	if (eigenval[*((int *)j1)] < eigenval[*((int *)j2)]) return 1;
	return eigenval[*((int *)j1)] > eigenval[*((int *)j2)] ? -1 : 0;
}

/*
 * Diagonalize a real symmetric matrix.
 *
 * `keepin' is size-x-size and will NOT be destroyed.
 * `out' is size-x-size and the ROWS will be set to the eigenvectors of `in'.
 * `vals' is size-length and will be set to the eigenvalues.
 * the eigens are dsorted by decreasing eigenvalue.
 */
void make_eigen(double *keepin, double *out, double *vals, int size)
{
	double	*work, *in;
	int	i, j, k, nrot;

	in = (double *)mustmalloc(size * size * sizeof(double), "Alloc IN");
	work = (double *)mustmalloc(size * size * sizeof(double), "Alloc WORK");
	memcpy(in, keepin, size * size * sizeof(double));
	jacobi(in, size, eigenval, work, &nrot);
	fprintf(stderr, "%d jacobi rotations required\n", nrot);
	for (i = 0; i < size; i++)
		jix[i] = i;
	qsort(jix, size, sizeof(*jix), jixcomp);
	for (i = 0; i < size; i++, out += size) {
		k = jix[i];
		vals[i] = eigenval[k];
		for (j = 0; j < size; j++)
			out[j] = work[j * size + k];
	}
	free(in);
	free(work);
}

/*********************************/
#define ROTATE(a,i,j,k,l)		\
	g=a[i*n + j];			\
	h=a[k*n + l];			\
	a[i*n + j]=g-s*(h+g*tau);	\
	a[k*n + l]=h+s*(g-h*tau);

/*
 * Set v (resp d) to the eigenvectors (resp vales)
 *	of the nxn real sym matrix a;
 * Set *nrot to the number of jacobi rotations required.
 */
void jacobi(double *a, int n, double *d, double *v, int *nrot)
{
	int j, iq, ip, i;
	double tresh, theta, tau, t, sm, s, h, g, c;
	double	b[MAXDIM];
	double	z[MAXDIM];

	for (ip=0; ip<n; ip++) {
		for (iq=0; iq<n; iq++)
			v[ip*n + iq]=0.0;
		v[ip*n + ip]=1.0;
	}
	for (ip=0; ip<n; ip++) {
		b[ip]=d[ip]=a[ip*n + ip];
		z[ip]=0.0;
	}
	*nrot=0;
	for (i=1; i<=50; i++) {
		sm=0.0;
		for (ip=0; ip<n-1; ip++) {
			for (iq=ip+1; iq<n; iq++)
				sm += fabs(a[ip*n + iq]);
		}
		if (sm == 0.0) {
			return;
		}
		if (i < 4)
			tresh=0.2*sm/(n*n);
		else
			tresh=0.0;
		for (ip=0; ip<n-1; ip++) {
			for (iq=ip+1; iq<n; iq++) {
				g=100.0*fabs(a[ip*n + iq]);
				if (i > 4 && fabs(d[ip])+g
						== fabs(d[ip])
					&& fabs(d[iq])+g
						== fabs(d[iq]))
					a[ip*n + iq]=0.0;
				else if (fabs(a[ip*n + iq]) > tresh) {
					h=d[iq]-d[ip];
					if (fabs(h)+g == fabs(h))
						t=(a[ip*n + iq])/h;
					else {
						theta=0.5*h/(a[ip*n + iq]);
						t=1.0/(fabs(theta)
							+sqrt(1.0+theta*theta));
						if (theta < 0.0) t = -t;
					}
					c=1.0/sqrt(1+t*t);
					s=t*c;
					tau=s/(1.0+c);
					h=t*a[ip*n + iq];
					z[ip] -= h;
					z[iq] += h;
					d[ip] -= h;
					d[iq] += h;
					a[ip*n + iq]=0.0;
					for (j=0; j<=ip-1; j++) {
						ROTATE(a, j, ip, j, iq)
					}
					for (j=ip+1; j<=iq-1; j++) {
						ROTATE(a, ip, j, j, iq)
					}
					for (j=iq+1; j<n; j++) {
						ROTATE(a, ip, j, iq, j)
					}
					for (j=0; j<n; j++) {
						ROTATE(v, j, ip, j, iq)
					}
					++(*nrot);
				}
			}
		}
		for (ip=0; ip<n; ip++) {
			b[ip] += z[ip];
			d[ip]=b[ip];
			z[ip]=0.0;
		}
	}
	fprintf(stderr, "Too many iterations in routine JACOBI\n");
	exit(EXIT_EXCEED);
}

/* Make v[] a random vector */
void mk_randvect(double *v, int size)
{
	while (size--)
		*v++ = SIMP_RPROB * 2 - 1;
}

/* Set B to the NxN matrix I - A where A is NxN */
void iminusa(double *a, double *b, int AWID)
{
	int	i, j;

	for (i = 0; i < AWID; i++)
	for (j = 0; j < AWID; j++)
		*b++ = (i == j) - *a++;
}

/* Here's another useful idiom; projector matrix */
/* Set B to the NxN matrix A^T (A A^T)^(-1) A where A is MxN */
void mk_projector(double *a, double *b, int AHEI, int AWID)
{
	double	*at, *prod;
	int	i, j, k;

	if (AWID < AHEI) {
		fprintf(stderr, "Mk_projector: size\n");
		exit(EXIT_PINCON);
	}
	at = (double *)mustmalloc(AHEI * AWID * sizeof(double), "Alloc AT");
	prod = (double *)mustmalloc(AHEI * AWID * sizeof(double), "Alloc PROD");
	transpose(a, at, AWID, AHEI);
	multiply(a, at, prod, AWID, AHEI, AHEI);
	if (do_invert(prod, b, AHEI) == 0) {
		fprintf(stderr, "Mk_projector: singularity\n");
		exit(EXIT_PINCON);
	}
	multiply(b, a, prod, AHEI, AHEI, AWID);
	multiply(at, prod, b, AHEI, AWID, AWID);
	free(at);
	free(prod);
}

/* Is `a' a projector? */
int is_projector(double *a, int AHEI)
{
	int	i;	
	double	*prod, devia;

	prod = (double *)mustmalloc(AHEI * AHEI * sizeof(double), "Alloc PROD");
	multiply(a, a, prod, AHEI, AHEI, AHEI);
	for (devia = i = 0; i < AHEI * AHEI; i++)
		devia += fabs(prod[i] - a[i]);
	free(prod);
	if (devia > 0.0000001) {
		fprintf(stderr, "Is_projector --> NO  Deviation = %f\n",
				devia);
		return 0;
	} else {
		return 1;
	}
}

/* Make `in[next]' ortho to first `next' rows of `in'; given that `in' is .. */
void nxt_ortho(double *in, int size, int next)
{
	double	*proj, *xwork, rvect[MAXDIM];

	xwork = (double *)mustmalloc(size * size * sizeof(double), "Alloc XWORK");
	proj = (double *)mustmalloc(size * size * sizeof(double), "Alloc PROJ");
	do {
		mk_randvect(rvect, size);
		transpose(in, xwork, size, next);
		multiply(xwork, in, proj, next, size, size);
		iminusa(proj, xwork, size);
		is_projector(xwork, size);
		multiply(xwork, rvect, in + size * next, size, size, 1);
		if (verbose == 9) {
			fprintf(stderr, "%d %d  ", size, next);
			aprint("Rvect", rvect, size, 1);
		}
	} while (fabs(ddotprod(in + size * next,
			in + size * next, size, 1)) < 0.01);
	normalize(in + size * next, size, 1);
	free(xwork);
	free(proj);
}

/* Make `in' orthogonal given that the first `already' rows already are */
void fill_ortho(double *in, int size, int already)
{
	while (already < size)
		nxt_ortho(in, size, already++);
}

/*********** END OF "MATH LIBRARY" ********/

/* Rest of file is standalone test driver */
#if	0
#define	TESTING
#endif
#ifdef	TESTING
#define	TSIZE	12
double	A[TSIZE][TSIZE];
double	B[TSIZE][TSIZE];
double	C[TSIZE][TSIZE];


int main(int argc, char **argv)
{
	int	i, j;

	for (i = 0; i < TSIZE; i++)
	for (j = 0; j < TSIZE; j++)
		A[i][j] = SIMP_RPROB * 999.0;
	aprint("In", A[0], TSIZE, TSIZE);

#if	0
	invert44(A[0], B[0]);
	aprint("Inverse", B[0], TSIZE, TSIZE);
	multiply(A[0], B[0], C[0], TSIZE, TSIZE, TSIZE);
	aprint("Prod", C[0], TSIZE, TSIZE);
#endif
#if	0
	fprintf(stderr, "Determ = %f\n", MDET4(A));
	fprintf(stderr, "Deter? = %f\n", determ(A[0], 4));
	exit(EXIT_GOOD);
#endif

	matinvert(A[0], B[0], TSIZE);
	aprint("Inverse", B[0], TSIZE, TSIZE);
	multiply(A[0], B[0], C[0], TSIZE, TSIZE, TSIZE);
	aprint("Prod", C[0], TSIZE, TSIZE);
	exit(EXIT_GOOD);
}
#endif

